# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
from typing import Any, Dict, List, Optional, Tuple, Union, overload
from forte.utils import create_import_error_msg

try:
    import torch
    from torch import nn
    from torch.nn import functional as F
except ImportError as e:
    raise ImportError(
        create_import_error_msg("torch", "models", "models")
    ) from e

try:
    import texar.torch as tx
except ImportError as e:
    raise ImportError(
        create_import_error_msg(
            "texar-pytorch", "models", "Texar model support"
        )
    ) from e


LSTMState = Tuple[torch.Tensor, torch.Tensor]


class CustomLSTMCell(tx.core.RNNCellBase[LSTMState]):
    # pylint: disable=super-init-not-called
    def __init__(self, input_size: int, hidden_size: int, dropout: float = 0.0):
        # skip super class constructor
        nn.Module.__init__(self)  # pylint: disable=non-parent-init-called

        self._input_size = input_size
        self._hidden_size = hidden_size

        self._dropout_rate = dropout
        self._dropout_mask: Optional[torch.Tensor] = None

        self._initial_cell = nn.Parameter(torch.Tensor(1, hidden_size))
        self._initial_hidden = nn.Parameter(torch.Tensor(1, hidden_size))
        self._projection = nn.Linear(input_size + hidden_size, 3 * hidden_size)

        self.reset_parameters()

    @property
    def input_size(self) -> int:
        return self._input_size

    @property
    def hidden_size(self) -> int:
        return self._hidden_size

    @staticmethod
    @torch.no_grad()
    def orthonormal_init(param: nn.Parameter, n_blocks: int):
        size0, size1 = param.size()
        size0 //= n_blocks
        size_min = min(size0, size1)
        init_values = []
        for _ in range(n_blocks):
            m1 = torch.randn(size0, size0, dtype=param.dtype)
            m2 = torch.randn(size1, size1, dtype=param.dtype)
            q1, r1 = torch.qr(m1)
            q2, r2 = torch.qr(m2)
            q1 *= torch.sign(torch.diag(r1))
            q2 *= torch.sign(torch.diag(r2))
            value = torch.mm(q1[:, :size_min], q2[:size_min, :])
            init_values.append(value)
        param.data = torch.cat(init_values, dim=0)

    def reset_parameters(self):
        self.orthonormal_init(self._projection.weight, 3)
        nn.init.xavier_normal_(self._initial_cell)
        nn.init.xavier_normal_(self._initial_hidden)

    def init_batch(self):
        self._dropout_mask = None

    def zero_state(self, batch_size: int) -> LSTMState:
        return (
            self._initial_hidden.expand(batch_size, -1),
            self._initial_cell.expand(batch_size, -1),
        )

    def forward(
        self, input: torch.Tensor, state: Optional[LSTMState] = None
    ) -> Tuple[torch.Tensor, LSTMState]:
        batch_size = input.size(0)
        if state is None:
            state = self.zero_state(batch_size)
        h, c = state
        if self.training and self._dropout_rate > 0.0:
            if self._dropout_mask is None:
                keep_prob = 1 - self._dropout_rate
                self._dropout_mask = input.new_zeros(
                    batch_size, self._hidden_size
                ).bernoulli_(keep_prob)
            h = h * self._dropout_mask
        concat_proj = self._projection(torch.cat([input, h], dim=1))
        i, g, o = torch.chunk(concat_proj, 3, dim=1)
        i = torch.sigmoid(i)
        new_c = (1 - i) * c + i * torch.tanh(g)
        new_h = torch.tanh(new_c) * torch.sigmoid(o)
        return new_h, (new_h, new_c)


class CustomBiLSTM(tx.modules.EncoderBase):
    def __init__(self, hparams=None):
        super().__init__(hparams)

        self.fw_cells = nn.ModuleList()
        self.bw_cells = nn.ModuleList()
        input_dim = self._hparams.input_dim
        hidden_dim = self._hparams.hidden_dim
        for _ in range(self._hparams.num_layers):
            fw_cell = CustomLSTMCell(input_dim, hidden_dim)
            bw_cell = CustomLSTMCell(input_dim, hidden_dim)
            self.fw_cells.append(fw_cell)
            self.bw_cells.append(bw_cell)
            input_dim = 2 * hidden_dim

        self.highway_layers = nn.ModuleList(
            [
                nn.Linear(hidden_dim * 2, hidden_dim * 2)
                for _ in range(self._hparams.num_layers - 1)
            ]
        )

        self.dropout = nn.Dropout(self._hparams.dropout)

    @staticmethod
    def default_hparams() -> Dict[str, Any]:
        return {
            "input_dim": 200,
            "hidden_dim": 200,
            "num_layers": 3,
            "dropout": 0.2,
        }

    def forward(
        self,
        inputs: torch.Tensor,
        sequence_length: Optional[torch.LongTensor] = None,
    ) -> torch.Tensor:
        for idx in range(self._hparams.num_layers):
            (
                fw_outputs,
                bw_outputs,
            ), _ = tx.utils.rnn.bidirectional_dynamic_rnn(
                self.fw_cells[idx], self.bw_cells[idx], inputs, sequence_length
            )
            outputs = torch.cat([fw_outputs, bw_outputs], dim=2)
            outputs = self.dropout(outputs)
            if idx > 0:
                gate = self.highway_layers[idx - 1](outputs)
                outputs = gate * outputs + (1 - gate) * inputs
            inputs = outputs
        return inputs


class CharCNN(tx.ModuleBase):
    __torch_device__: torch.device

    def __init__(self, char_vocab_size, hparams=None):
        super().__init__(hparams)

        self.char_embed = tx.modules.WordEmbedder(
            vocab_size=char_vocab_size,
            hparams={
                "dim": self._hparams.char_embed_size,
            },
        )
        self.cnn_kernels = nn.ModuleList(
            [
                nn.Conv1d(
                    self._hparams.char_embed_size,
                    self._hparams.filter_size,
                    kernel_size=width,
                )
                for width in self._hparams.filter_widths
            ]
        )
        self._max_filter_width = max(self._hparams.filter_widths)

    @staticmethod
    def default_hparams() -> Dict[str, Any]:
        return {
            "char_embed_size": 8,
            "filter_widths": [3, 4, 5],
            "filter_size": 50,
        }

    @property
    def output_size(self):
        return len(self._hparams.filter_widths) * self._hparams.filter_size

    @property
    def _device(self) -> torch.device:
        if not hasattr(self, "__torch_device__"):
            self.__torch_device__ = next(self.parameters()).device
        return self.__torch_device__

    def forward(
        self, batch: torch.Tensor, masks: List[torch.Tensor]
    ) -> torch.Tensor:
        r"""Construct character-level representations of the words.

        Args:
            batch: A batch of padded words having the shape
            `(batch_size, max_length, max_word_length)`, where:

            - `batch_size` is the number of sentences in this batch
            - `max_length` is equal to ``max(len(s) for s in words)``, i.e.
              length of the longest sentence in the batch.
            - `max_word_length` is max length of word for all words in this
              batch. All words

            For sentences shorter than `max_length`, the extra positions are
            padded with zero vectors.
            For words shorter than `max_word_length`, the extra positions are
            padded with zero vectors.

            masks: A list of masks corresponding to each dimension in the batch.
            ``masks[0]`` represents the mask for each sentence. It has
            the shape `(batch_size, max_length)`
            ``masks[1]`` represents the mask for each word in each sentence. It
            has the shape `(batch_size, max_length, max_word_length)`

        Returns:
            A tensor of shape `(batch_size, max_length, output_size)`, where

            - `batch_size` is equal to ``len(words)``.
            - `max_length` is equal to ``max(len(s) for s in words)``, i.e.
              length of the longest sentence in the batch.
            - `output_size` is equal to ``self.output_size``.

            For sentences shorter than `max_length`, the extra positions are
            padded with zero vectors.
        """
        batch_size, max_length, max_word_length = batch.size()
        sentence_mask = masks[0]
        sent_lengths = torch.sum(sentence_mask, dim=1).tolist()

        # indices: (len(all_words), max_word_length)
        indices = torch.cat(
            [batch[i][:l] for i, l in enumerate(sent_lengths)], dim=0
        )

        # TODO: is embed.shape[1] batch_size * max_length or len(all_words) ?
        # embed: (batch_size * max_length, char_embed_dim, indices)
        embed = self.char_embed(indices).transpose(1, 2)
        if max_word_length < self._max_filter_width:
            pad_length = self._max_filter_width - max_word_length
            # TODO: Probably Train preprocessor can help do this padding
            embed = torch.cat(
                [embed, embed.new_zeros(*embed.size()[:2], pad_length)], dim=2
            )
        kernel_outputs = [kernel(embed) for kernel in self.cnn_kernels]
        cnn_output = torch.cat(
            [torch.max(out, dim=2)[0] for out in kernel_outputs], dim=1
        )

        sent_cnn_outputs = torch.split(cnn_output, sent_lengths, dim=0)
        output = cnn_output.new_zeros(
            batch_size, max_length, cnn_output.size(1)
        )
        for idx, sent_output in enumerate(sent_cnn_outputs):
            output[idx, : sent_lengths[idx]] = sent_output
        return output


def sum_list(xs: List[torch.Tensor]) -> torch.Tensor:
    if len(xs) == 1:
        return xs[0]
    mid = len(xs) // 2
    return sum_list(xs[:mid]) + sum_list(xs[mid:])


# pylint: disable=unused-argument,function-redefined
@overload
def batch_gather(tensors: torch.Tensor, index: torch.Tensor) -> torch.Tensor:
    ...


@overload
def batch_gather(  # type: ignore
    tensors: List[torch.Tensor], index: torch.Tensor
) -> List[torch.Tensor]:
    ...


def batch_gather(tensors, index):
    batch_size, query_size = index.size()
    value_size = None

    def _get_value_size_fn(tensor):
        nonlocal value_size
        if value_size is not None or not isinstance(tensor, torch.Tensor):
            return
        value_size = tensor.size(1)

    tx.utils.map_structure(_get_value_size_fn, tensors)
    # assert torch.max(index).item() < value_size
    offsets = torch.arange(batch_size, device=index.device)
    offsets = (offsets * value_size).repeat_interleave(query_size)
    index = index.contiguous().view(-1) + offsets

    def _gather_fn(tensor):
        if not isinstance(tensor, torch.Tensor):
            return tensor
        tensor = tensor.contiguous().view(-1, *tensor.size()[2:])
        values = torch.index_select(tensor, dim=0, index=index)
        values = values.view(batch_size, -1, *values.size()[1:])
        return values

    gathered_tensors = tx.utils.map_structure(_gather_fn, tensors)
    return gathered_tensors


# pylint: enable=unused-argument,function-redefined


class MLP(tx.ModuleBase):
    def __init__(self, hparams=None):
        super().__init__(hparams)

        input_size = self._hparams.input_size
        layers = []
        dropout = self._hparams.dropout_rate
        if self._hparams.activation is not None:
            activation = tx.core.get_layer({"type": self._hparams.activation})
        else:
            activation = None
        for has_bias, output_size in zip(
            self._hparams.has_bias, self._hparams.hidden_size
        ):
            layers.append(nn.Linear(input_size, output_size, bias=has_bias))
            input_size = output_size
            if activation is not None:
                layers.append(activation)
            if dropout > 0.0:
                layers.append(nn.Dropout(dropout, inplace=True))
        output_size = self._hparams.output_size
        layers.append(nn.Linear(input_size, output_size, bias=False))
        self.layers = nn.Sequential(*layers)

    @staticmethod
    def default_hparams() -> Dict[str, Any]:
        return {
            "input_size": 300,
            "num_layers": 2,
            "has_bias": [True, True],
            "hidden_size": [150, 150],
            "activation": None,
            "output_size": 1,
            "dropout_rate": 0.0,
        }

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return self.layers(input)


class ConcatInputMLP(tx.ModuleBase):
    _cached_inputs: List[torch.Tensor]

    def __init__(self, hparams=None):
        super().__init__(hparams)

        mlp_hparams = hparams.copy()
        mlp_hparams["input_size"] = sum(self._hparams.input_sizes)
        del mlp_hparams["input_sizes"]
        self.mlp = MLP(mlp_hparams)

    @staticmethod
    def default_hparams() -> Dict[str, Any]:
        return {
            "input_sizes": [150, 150],
            "num_layers": 2,
            "has_bias": [True, True],
            "hidden_size": [150, 150],
            "output_size": 1,
            "dropout_rate": 0.0,
            "activation": None,
        }

    @contextmanager
    def cache_results(self, inputs: List[torch.Tensor]):
        self._cached_inputs = inputs
        yield
        del self._cached_inputs

    def forward(
        self,
        inputs: List[
            Union[torch.Tensor, List[Tuple[torch.Tensor, torch.Tensor]]]
        ],
    ) -> torch.Tensor:
        """

        Args:
            inputs: List of 2D or 3D Tensors (or list of `(index, float)`
                tuples), representing indices for each concat'd part of the
                input tensor. First dimension is the batch dimension.

                - If an element of the list is another list, each inner list
                  element must be a tuple of `(index, coef)`, where `index` is
                  used for indexing, and all indexed tensors will be summed up,
                  weighted by the tensor `coef`.
                - If the tensor is 2D, it must be a LongTensor.
                - If the tensor is 3D, it must be a FloatTensor representing the
                  soft-IDs (weights for weighted average).

        Returns:
        """
        parts = []
        for index, cache in zip(inputs, self._cached_inputs):
            if isinstance(index, list):
                raise NotImplementedError
            if index.dim() == 3:
                if cache.dim() == 2:
                    # [batch, query_size, value_size] x [value_size, dim]
                    # Make it 3D so we can perform batched MM.
                    cache = cache.expand(index.size(0), *cache.size())
                # [batch, query_size, value_size] x [batch, value_size, dim]
                values = torch.bmm(index, cache)
            elif cache.dim() == 3:
                # Batch-specific values. Add batch offset to indices for lookup.
                values = batch_gather(cache, index)
            else:
                values = F.embedding(index, cache)
            parts.append(values)
        output = self.mlp(torch.cat(parts, dim=-1))
        return output



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import gzip
import pickle
import uuid
from abc import abstractmethod
from pathlib import Path
from typing import (
    List,
    Optional,
    Set,
    Type,
    TypeVar,
    Union,
    Iterator,
    Dict,
    Tuple,
    Any,
    Iterable,
)


import jsonpickle

from forte.common import ProcessExecutionException, EntryNotFoundError
from forte.data.container import EntryContainer
from forte.data.index import BaseIndex
from forte.data.ontology.core import Entry, EntryType, GroupType, LinkType
from forte.version import PACK_VERSION, DEFAULT_PACK_VERSION

__all__ = ["BasePack", "BaseMeta", "PackType"]


class BaseMeta:
    r"""Basic Meta information for both :class:`~forte.data.data_pack.DataPack`
    and :class:`~forte.data.multi_pack.MultiPack`.

    Args:
        pack_name:  An name to identify the data pack, which is helpful in
           situation like serialization. It is suggested that the packs should
           have different doc ids.
    Attributes:
        record: Initialized as a dictionary. This is not a required field.
            The key of the record should be the entry type and values should
            be attributes of the entry type. All the information would be used
            for consistency checking purpose if the pipeline is initialized with
            `enforce_consistency=True`.
    """

    def __init__(self, pack_name: Optional[str] = None):
        self.pack_name: Optional[str] = pack_name
        self._pack_id: int = uuid.uuid4().int
        self.record: Dict[str, Set[str]] = {}

    def __getstate__(self):
        state = self.__dict__.copy()
        return state

    def __setstate__(self, state):
        """
        Re-obtain the pack manager during deserialization.
        Args:
            state:

        Returns:

        """
        self.__dict__.update(state)

    @property
    def pack_id(self) -> int:
        return self._pack_id


class BasePack(EntryContainer[EntryType, LinkType, GroupType]):
    r"""The base class of :class:`~forte.data.data_pack.DataPack` and
    :class:`~forte.data.multi_pack.MultiPack`.

    Args:
        pack_name: a string name of the pack.

    """

    # pylint: disable=too-many-public-methods
    def __init__(self, pack_name: Optional[str] = None):
        super().__init__()
        self.links: List[LinkType] = []
        self.groups: List[GroupType] = []
        self.pack_version: str = PACK_VERSION

        self._meta: BaseMeta = self._init_meta(pack_name)
        self._index: BaseIndex = BaseIndex()

        self.__control_component: Optional[str] = None

        # This Dict maintains a mapping from entry's tid to the Entry object
        # itself and the component name associated with the entry.
        # The component name is used for tracking the "creator" of this entry.
        self._pending_entries: Dict[int, Tuple[Entry, Optional[str]]] = {}

    def __getstate__(self):
        state = self.__dict__.copy()
        state.pop("_index")
        state.pop("_pending_entries")
        state.pop("_BasePack__control_component")
        return state

    def __setstate__(self, state):
        super().__setstate__(state)
        if "meta" in self.__dict__:
            self._meta = self.__dict__.pop("meta")
        self.__control_component = None
        self._pending_entries = {}

    @abstractmethod
    def _init_meta(self, pack_name: Optional[str] = None) -> BaseMeta:
        raise NotImplementedError

    def set_meta(self, **kwargs):
        for k, v in kwargs.items():
            if not hasattr(self._meta, k):
                raise AttributeError(f"Meta has no attribute named {k}")
            setattr(self._meta, k, v)

    @property
    def pack_id(self):
        return self._meta.pack_id

    @abstractmethod
    def __iter__(self) -> Iterator[EntryType]:
        raise NotImplementedError

    def __del__(self):
        if len(self._pending_entries) > 0:
            raise ProcessExecutionException(
                f"There are {len(self._pending_entries)} "
                f"entries not added to the index correctly."
            )

    @property
    def pack_name(self):
        return self._meta.pack_name

    @pack_name.setter
    def pack_name(self, pack_name: str):
        """
        Update the pack name of this pack.

        Args:
            pack_name: The new doc id.

        Returns:

        """
        self._meta.pack_name = pack_name

    @classmethod
    def _deserialize(
        cls,
        data_source: Union[Path, str],
        serialize_method: str = "jsonpickle",
        zip_pack: bool = False,
    ) -> "PackType":
        """
        This function should deserialize a Pack from a string. The
        implementation should decide the specific pack type.

        Args:
            data_source: The data path containing pack data. The content
              of the data could be string or bytes depending on the method of
              serialization.
            serialize_method: The method used to serialize the data, this
              should be the same as how serialization is done. The current
              options are `jsonpickle` and `pickle`. The default method
              is `jsonpickle`.
            zip_pack: Boolean value indicating whether the input source is
              zipped.

        Returns:
            An pack object deserialized from the data.
        """
        _open = gzip.open if zip_pack else open

        if serialize_method == "jsonpickle":
            with _open(data_source, mode="rt") as f:  # type: ignore
                pack = cls.from_string(f.read())
        else:
            with _open(data_source, mode="rb") as f:  # type: ignore
                pack = pickle.load(f)

            if not hasattr(pack, "pack_version"):
                pack.pack_version = DEFAULT_PACK_VERSION

        return pack  # type: ignore

    @classmethod
    def from_string(cls, data_content: str) -> "BasePack":
        pack = jsonpickle.decode(data_content)
        if not hasattr(pack, "pack_version"):
            pack.pack_version = DEFAULT_PACK_VERSION

        return pack

    @abstractmethod
    def delete_entry(self, entry: EntryType):
        r"""Remove the entry from the pack.

        Args:
            entry: The entry to be removed.

        Returns:
            None
        """
        raise NotImplementedError

    def add_entry(
        self, entry: Entry, component_name: Optional[str] = None
    ) -> EntryType:
        r"""Add an :class:`~forte.data.ontology.core.Entry` object to the
        :class:`~forte.data.base_pack.BasePack` object. Allow duplicate entries in a pack.

        Args:
            entry: An :class:`~forte.data.ontology.core.Entry`
                object to be added to the pack.
            component_name: A name to record that the entry is created by
             this component.

        Returns:
            The input entry itself
        """
        # When added to the pack, make a record.
        self.record_entry(entry, component_name)
        # TODO: Returning the entry itself may not be helpful.
        return self._add_entry(entry)

    @abstractmethod
    def _add_entry(self, entry: Entry) -> EntryType:
        r"""Add an :class:`~forte.data.ontology.core.Entry` object to the
        :class:`~forte.data.base_pack.BasePack` object. Allow duplicate entries in a pack.

        Args:
            entry: An :class:`~forte.data.ontology.core.Entry`
                object to be added to the pack.

        Returns:
            The input entry itself
        """
        raise NotImplementedError

    def add_all_remaining_entries(self, component: Optional[str] = None):
        """
        Calling this function will add the entries that are not added to the
        pack manually.

        Args:
            component: Overwrite the component record with this.

        Returns:
            None
        """
        for entry, c in list(self._pending_entries.values()):
            c_ = component if component else c
            self.add_entry(entry, c_)
        self._pending_entries.clear()

    def to_string(
        self,
        drop_record: Optional[bool] = False,
        json_method: str = "jsonpickle",
        indent: Optional[int] = None,
    ) -> str:
        """
        Return the string representation (json encoded) of this method.

        Args:
            drop_record: Whether to drop the creation records, default is False.
            json_method: What method is used to convert data pack to json.
              Only supports `json_pickle` for now. Default value is
              `json_pickle`.
            indent: The indent used for json string.

        Returns: String representation of the data pack.
        """
        if drop_record:
            self._creation_records.clear()
            self._field_records.clear()
        if json_method == "jsonpickle":
            return jsonpickle.encode(self, unpicklable=True, indent=indent)
        else:
            raise ValueError(f"Unsupported JSON method {json_method}.")

    def serialize(
        self,
        output_path: Union[str, Path],
        zip_pack: bool = False,
        drop_record: bool = False,
        serialize_method: str = "jsonpickle",
        indent: Optional[int] = None,
    ):
        r"""
        Serializes the data pack to the provided path. The output of this
        function depends on the serialization method chosen.

        Args:
            output_path: The path to write data to.
            zip_pack: Whether to compress the result with `gzip`.
            drop_record: Whether to drop the creation records, default is False.
            serialize_method: The method used to serialize the data. Currently
              supports `jsonpickle` (outputs str) and Python's built-in
              `pickle` (outputs bytes).
            indent: Whether to indent the file if written as JSON.

        Returns: Results of serialization.
        """
        if zip_pack:
            _open = gzip.open
        else:
            _open = open  # type:ignore

        if drop_record:
            self._creation_records.clear()
            self._field_records.clear()

        if serialize_method == "pickle":
            with _open(output_path, mode="wb") as pickle_out:
                pickle.dump(self, pickle_out)
        elif serialize_method == "jsonpickle":
            with _open(output_path, mode="wt", encoding="utf-8") as json_out:
                json_out.write(
                    self.to_string(drop_record, "jsonpickle", indent=indent)
                )
        else:
            raise NotImplementedError(
                f"Unsupported serialization method {serialize_method}"
            )

    def view(self):
        return copy.deepcopy(self)

    def set_control_component(self, component: str):
        """
        Record the current component that is taking control of this pack.

        Args:
            component: The component that is going to take control

        Returns:

        """
        self.__control_component = component

    def record_entry(self, entry: Entry, component_name: Optional[str] = None):
        c = component_name

        if c is None:
            # Use the auto-inferred control component.
            c = self.__control_component

        if c is not None:
            try:
                self._creation_records[c].add(entry.tid)
            except KeyError:
                self._creation_records[c] = {entry.tid}

    def record_field(self, entry_id: int, field_name: str):
        """
        Record who modifies the entry, will be called
        in :class:`~forte.data.ontology.core.Entry`

        Args:
            entry_id: The id of the entry.
            field_name: The name of the field modified.

        Returns:

        """
        c = self.__control_component

        if c is not None:
            try:
                self._field_records[c].add((entry_id, field_name))
            except KeyError:
                self._field_records[c] = {(entry_id, field_name)}

    def on_entry_creation(
        self, entry: Entry, component_name: Optional[str] = None
    ):
        """
        Call this when adding a new entry, will be called
        in :class:`~forte.data.ontology.core.Entry` when
        its `__init__` function is called.

        Args:
            entry: The entry to be added.
            component_name: A name to record that the entry is created by
             this component.

        Returns:

        """
        c = component_name

        if c is None:
            # Use the auto-inferred control component.
            c = self.__control_component

        # Record that this entry hasn't been added to the index yet.
        self._pending_entries[entry.tid] = entry, c

    def regret_creation(self, entry: EntryType):
        """
        Will remove the entry from the pending entries internal state of the
        pack.

        Args:
            entry: The entry that we would not add the the pack anymore.

        Returns:

        """
        self._pending_entries.pop(entry.tid)

    # TODO: how to make this return the precise type here?
    def get_entry(self, tid: int) -> EntryType:
        r"""Look up the entry_index with key ``ptr``. Specific implementation
        depends on the actual class."""
        entry: EntryType = self._index.get_entry(tid)
        if entry is None:
            raise KeyError(
                f"There is no entry with tid '{tid}'' in this datapack"
            )
        return entry

    @abstractmethod
    def get_data(
        self, context_type, request, skip_k
    ) -> Iterator[Dict[str, Any]]:
        raise NotImplementedError

    @abstractmethod
    def get(
        self, entry_type: Union[str, Type[EntryType]], **kwargs
    ) -> Iterator[EntryType]:
        """
        Implementation of this method should provide to obtain the entries in
        entry ordering. If there are orders defined between the entries, they
        should be used first. Otherwise, the insertion order should be
        used (FIFO).

        Args:
            entry_type: The type of the entry to obtain.

        Returns:
            An iterator of the entries matching the provided arguments.
        """
        raise NotImplementedError

    def get_single(self, entry_type: Union[str, Type[EntryType]]) -> EntryType:
        r"""Take a single entry of type
        :attr:`~forte.data.data_pack.DataPack.entry_type` from this data
        pack. This is useful when the target entry type appears only one
        time in the :class:`~forte.data.data_pack.DataPack` for e.g., a Document entry. Or you just
        intended to take the first one.

        Args:
            entry_type: The entry type to be retrieved.

        Returns:
            A single data entry.
        """
        for a in self.get(entry_type):
            return a

        raise EntryNotFoundError(
            f"The entry {entry_type} is not found in the provided pack."
        )

    def get_ids_by_creator(self, component: str) -> Set[int]:
        r"""
        Look up the component_index with key `component`. This will return
        the entry ids that are created by the `component`

        Args:
            component: The component (creator) to find ids for.

        Returns:
            A set of entry ids that are created by the component.
        """
        entry_set: Set[int] = self._creation_records[component]
        return entry_set

    def is_created_by(
        self, entry: Entry, components: Union[str, Iterable[str]]
    ) -> bool:
        """
        Check if the entry is created by any of the provided components.

        Args:
            entry: The entry to check.
            components: The list of component names.

        Returns:
            True if the entry is created by the component, False otherwise.
        """
        if isinstance(components, str):
            components = [components]

        for c in components:
            if entry.tid in self._creation_records[c]:
                break
        else:
            # The entry not created by any of these components.
            return False

        return True

    def get_entries_from(self, component: str) -> Set[EntryType]:
        """
        Look up all entries from the `component` as a unordered set

        Args:
            component: The component (creator) to get the entries. It is
                normally the full qualified name of the creator class, but it
                may also be customized based on the implementation.

        Returns:
            The set of entry ids that are created by the input component.
        """
        return {
            self.get_entry(tid) for tid in self.get_ids_by_creator(component)
        }

    def get_ids_from(self, components: List[str]) -> Set[int]:
        """
        Look up entries using a list of components (creators). This will find
        each creator iteratively and combine the result.

        Args:
            components: The list of components to find.

        Returns:
            The list of entry ids that are created from these components.
        """
        valid_component_id: Set[int] = set()
        for component in components:
            valid_component_id |= self.get_ids_by_creator(component)
        return valid_component_id

    def _expand_to_sub_types(self, entry_type: Type[EntryType]) -> Set[Type]:
        """
        Return all the types and the sub types that inherit from the provided
        type.

        Args:
            entry_type: The provided type to search for entry.

        Returns:
            A set of all the sub-types extending the provided type, including
            the input ``entry_type`` itself.
        """
        all_types: Set[Type] = set()
        for data_type in self._index.indexed_types():
            if issubclass(data_type, entry_type):
                all_types.add(data_type)
        return all_types

    def get_entries_of(
        self, entry_type: Type[EntryType], exclude_sub_types=False
    ) -> Iterator[EntryType]:
        """
        Return all entries of this particular type without orders. If you
        need to get the annotations based on the entry ordering,
        use :meth:`forte.data.base_pack.BasePack.get`.

        Args:
            entry_type: The type of the entry you are looking for.
            exclude_sub_types: Whether to ignore the inherited sub type
                of the provided `entry_type`. Default is True.

        Returns:
            An iterator of the entries matching the type constraint.
        """
        if exclude_sub_types:
            for tid in self._index.query_by_type(entry_type):
                yield self.get_entry(tid)
        else:
            for tid in self._index.query_by_type_subtype(entry_type):
                yield self.get_entry(tid)

    @classmethod
    @abstractmethod
    def validate_link(cls, entry: EntryType) -> bool:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def validate_group(cls, entry: EntryType) -> bool:
        raise NotImplementedError

    def get_links_from_node(
        self, node: Union[int, EntryType], as_parent: bool
    ) -> List[LinkType]:
        links: List[LinkType] = []
        if isinstance(node, Entry):
            tid = node.tid
            if tid is None:
                raise ValueError(
                    "The requested node has no tid. "
                    "Have you add this entry into the datapack?"
                )
        elif isinstance(node, int):
            tid = node
        else:
            raise TypeError(
                "Can only get group via entry id (int) or the "
                "group object itself (Entry)."
            )

        if not self._index.link_index_on:
            self._index.build_link_index(self.links)

        for tid in self._index.link_index(tid, as_parent=as_parent):
            entry: EntryType = self.get_entry(tid)
            if self.validate_link(entry):
                links.append(entry)  # type: ignore
        return links

    def get_links_by_parent(
        self, parent: Union[int, EntryType]
    ) -> List[LinkType]:
        return self.get_links_from_node(parent, True)

    def get_links_by_child(
        self, child: Union[int, EntryType]
    ) -> List[LinkType]:
        return self.get_links_from_node(child, False)

    def get_groups_by_member(
        self, member: Union[int, EntryType]
    ) -> Set[GroupType]:
        groups: Set[GroupType] = set()
        if isinstance(member, Entry):
            tid = member.tid
            if tid is None:
                raise ValueError(
                    "Argument member has no tid. "
                    "Have you add this entry into the datapack?"
                )
        elif isinstance(member, int):
            tid = member
        else:
            raise TypeError(
                "Can only get group via entry id (int) or the "
                "group object itself (Entry)."
            )

        if not self._index.group_index_on:
            self._index.build_group_index(self.groups)

        for tid in self._index.group_index(tid):
            entry: EntryType = self.get_entry(tid)
            if self.validate_group(entry):
                groups.add(entry)  # type: ignore
        return groups


PackType = TypeVar("PackType", bound=BasePack)



import os
from typing import List, Set

from string import Template

from forte.data.base_pack import PackType
from forte.data.data_pack import DataPack
from forte.data.multi_pack import MultiPack
from forte.data.ontology import top
from forte.data.ontology import utils


class SchemaKeywords:
    ontology_name = "name"
    imports = "imports"
    prefixes = "additional_prefixes"
    definitions = "definitions"
    parent_entry = "parent_entry"
    entry_name = "entry_name"
    description = "description"
    attributes = "attributes"
    attribute_name = "name"
    attribute_type = "type"
    parent_type = "parent_type"
    child_type = "child_type"
    member_type = "member_type"
    default_value = "default"
    element_type = "item_type"
    dict_key_type = "key_type"
    dict_value_type = "value_type"
    ndarray_dtype = "ndarray_dtype"
    ndarray_shape = "ndarray_shape"


# Some names are used as properties by the core types, they should not be
# reused by attributes.
RESERVED_ATTRIBUTE_NAMES: Set[str] = {
    "text",
    "span",
    "begin",
    "end",
    "index_key",
    "parent",
    "child",
    "embedding",
    "tid",
    "pack",
    "pack_id",
    "index_key",
}

REQUIRED_IMPORTS: List[str] = ["typing"]

TOP_MOST_MODULE_NAME = "forte.data.ontology.core"

DEFAULT_CONSTRAINTS_KEYS = {
    "BaseLink": {
        SchemaKeywords.parent_type: "ParentType",
        SchemaKeywords.child_type: "ChildType",
    },
    "BaseGroup": {SchemaKeywords.member_type: "MemberType"},
}

AUTO_GEN_SIGNATURE = "***automatically_generated***"
AUTO_GEN_FILENAME = ".generated"
AUTO_DEL_FILENAME = ".deleted"

SOURCE_JSON_PFX = "***source json:"
SOURCE_JSON_SFX = "***"
SOURCE_JSON_TEMP = Template(f"{SOURCE_JSON_PFX}$file_path{SOURCE_JSON_SFX}")


def get_ignore_error_lines(json_filepath: str) -> List[str]:
    source_json_sign = SOURCE_JSON_TEMP.substitute(file_path=json_filepath)
    return [
        f"# {AUTO_GEN_SIGNATURE}",
        f"# {source_json_sign}",
        "# flake8: noqa",
        "# mypy: ignore-errors",
        "# pylint: skip-file",
    ]


DEFAULT_PREFIX = "ft.onto"

SUPPORTED_PRIMITIVES = {"int", "float", "str", "bool"}
NON_COMPOSITES = {key: key for key in SUPPORTED_PRIMITIVES}
COMPOSITES = {"List", "Dict", "NdArray"}

ALL_INBUILT_TYPES = set(list(NON_COMPOSITES.keys()) + list(COMPOSITES))


def file_header(desc_str, ontology_name):
    desc_str = "" if desc_str is None else desc_str.strip()
    desc_str = desc_str + "\n" if desc_str else ""
    return (
        f"{desc_str}"
        f"Automatically generated ontology {ontology_name}. "
        f"Do not change manually."
    )


def class_name(clazz):
    return ".".join((clazz.__module__, clazz.__name__))


SINGLE_PACK_CLASSES = [class_name(clazz) for clazz in top.SinglePackEntries]
MULTI_PACK_CLASSES = [class_name(clazz) for clazz in top.MultiPackEntries]

major_version, minor_version = utils.get_python_version()
if major_version >= 3 and minor_version >= 7:
    PACK_TYPE_CLASS_NAME = class_name(PackType)
else:
    # bug in python < 3.7
    # returns    => typing.TypeVar('').__module__ == 'typing' (wrong)
    # instead of => typing.TypeVar('').__module__ == 'forte.data.base_pack'
    PACK_TYPE_CLASS_NAME = "forte.data.base_pack.PackType"


def hardcoded_pack_map(clazz):
    if clazz in SINGLE_PACK_CLASSES:
        return class_name(DataPack)
    elif clazz in MULTI_PACK_CLASSES:
        return class_name(MultiPack)
    else:
        # When not found, return the default.
        return PACK_TYPE_CLASS_NAME


class Config:
    indent: int = 4
    line_break: str = os.linesep



# ***automatically_generated***
# ***source json:forte/ontology_specs/medical.json***
# flake8: noqa
# mypy: ignore-errors
# pylint: skip-file
"""
Automatically generated ontology medical. Do not change manually.
"""

from dataclasses import dataclass
from forte.data.data_pack import DataPack
from forte.data.ontology.core import FList
from forte.data.ontology.top import Generics
from ft.onto.base_ontology import EntityMention
from typing import List
from typing import Optional

__all__ = [
    "UMLSConceptLink",
    "MedicalEntityMention",
]


@dataclass
class UMLSConceptLink(Generics):
    """
    A umls concept entity, used to represent basic information of a umls concept
    Attributes:
        cui (Optional[str]):
        name (Optional[str]):
        definition (Optional[str]):
        tuis (List[str]):
        aliases (List[str]):
        score (Optional[str]):
    """

    cui: Optional[str]
    name: Optional[str]
    definition: Optional[str]
    tuis: List[str]
    aliases: List[str]
    score: Optional[str]

    def __init__(self, pack: DataPack):
        super().__init__(pack)
        self.cui: Optional[str] = None
        self.name: Optional[str] = None
        self.definition: Optional[str] = None
        self.tuis: List[str] = []
        self.aliases: List[str] = []
        self.score: Optional[str] = None


@dataclass
class MedicalEntityMention(EntityMention):
    """
    A span based annotation class MedicalEntityMention, used to represent an Entity Mention in medical domain
    Attributes:
        umls_link (Optional[str]):
        umls_entities (FList[UMLSConceptLink]):
    """

    umls_link: Optional[str]
    umls_entities: FList[UMLSConceptLink]

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)
        self.umls_link: Optional[str] = None
        self.umls_entities: FList[UMLSConceptLink] = FList(self)



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This file wraps a machine translation model.
It could be used for back translation.
For simplicity, the model is not wrapped as a processor.
"""
from typing import List
from abc import abstractmethod
from forte.utils import create_import_error_msg

__all__ = [
    "MachineTranslator",
    "MarianMachineTranslator",
]


class MachineTranslator:
    r"""
    This class is a wrapper for machine translation models.

    Args:
        src_lang: The source language.
        tgt_lang: The target language.
        device: "cuda" for gpu, "cpu" otherwise.
    """

    def __init__(self, src_lang: str, tgt_lang: str, device: str):
        self.src_lang: str = src_lang
        self.tgt_lang: str = tgt_lang
        self.device = device

    @abstractmethod
    def translate(self, src_text: str) -> str:
        r"""
        This function translates the input text into target language.

        Args:
            src_text: The input text in source language.
        Returns:
            The output text in target language.
        """
        raise NotImplementedError


class MarianMachineTranslator(MachineTranslator):
    r"""
    This class is a wrapper for the Marian Machine Translator
    (https://huggingface.co/transformers/model_doc/marian.html).
    Please refer to their doc for supported languages.
    """

    def __init__(
        self, src_lang: str = "en", tgt_lang: str = "fr", device: str = "cpu"
    ):
        super().__init__(src_lang, tgt_lang, device)
        self.model_name = "Helsinki-NLP/opus-mt-{src}-{tgt}".format(
            src=src_lang, tgt=tgt_lang
        )
        try:
            from transformers import (  # pylint:disable=import-outside-toplevel
                MarianMTModel,
                MarianTokenizer,
            )
        except ImportError as err:
            raise ImportError(
                create_import_error_msg(
                    "transformers", "data_aug", "Machine Translator"
                )
            ) from err
        self.tokenizer = MarianTokenizer.from_pretrained(self.model_name)
        self.model = MarianMTModel.from_pretrained(self.model_name)
        self.model = self.model.to(self.device)

    def translate(self, src_text: str) -> str:
        translated: List[str] = self.model.generate(
            # TODO: Should not use prepare_seq2seq_batch for deprecation
            **self.tokenizer.prepare_seq2seq_batch(
                # Have to use explicitly call `convert_to_tensors` to make
                # this line work in both transformers 3 and 4, probably won't
                # work in 5.
                [src_text]
            )
            .convert_to_tensors("pt")
            .to(self.device)
        )

        tgt_texts: List[str] = [
            self.tokenizer.decode(t, skip_special_tokens=True)
            for t in translated
        ]
        return tgt_texts[0]



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Forte Span module.
"""
from functools import total_ordering

__all__ = [
    "Span",
]


@total_ordering
class Span:
    r"""A class recording the span of annotations. :class:`Span` objects can
    be totally ordered according to their ``begin`` as
    the first sort key and ``end`` as the second sort
    key.

    Args:
        begin: The offset of the first character in the span.
        end: The offset of the last character in the span + 1. So the
            span is a left-closed and right-open interval ``[begin, end)``.
    """

    def __init__(self, begin: int, end: int):
        if not isinstance(begin, int) or not isinstance(end, int):
            raise ValueError(
                f"Begin and End for an annotation must be integer, "
                f"got {begin}:{type(begin)} and {end}:{type(end)}"
            )

        if begin > end:
            raise ValueError(
                f"The begin {begin} of span is greater than the end {end}"
            )

        if begin < 0:
            raise ValueError("The begin cannot be negative.")

        self.begin = begin
        self.end = end

    def __repr__(self):
        return f"({self.begin}, {self.end})"

    def __lt__(self, other):
        if self.begin == other.begin:
            return self.end < other.end
        return self.begin < other.begin

    def __eq__(self, other):
        return (self.begin, self.end) == (other.begin, other.end)

    def __hash__(self):
        return hash((self.begin, self.end))



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
In this module, we provide a few readers to help reading DBpedia processed
 Wikipedia dumps. To use the readers here, the DBpedia full text datasets
 are needed: https://wiki.dbpedia.org/downloads-2016-10#p10608-2
"""
import csv
import logging
import os
from collections import defaultdict
from typing import (
    Iterator,
    Dict,
    List,
    Tuple,
    TextIO,
    Any,
    DefaultDict,
    Optional,
)


from smart_open import open

from forte.common import Resources
from forte.common.configuration import Config
from forte.common.exception import ResourceError
from forte.data.base_reader import PackReader
from forte.data.data_pack import DataPack
from forte.datasets.wikipedia.dbpedia.db_utils import (
    NIFParser,
    get_resource_attribute,
    get_resource_name,
    get_resource_fragment,
    print_progress,
    ContextGroupedNIFReader,
    state_type,
)
from forte.processors.base import PackWriter
from forte.utils import create_import_error_msg
from ft.onto.wikipedia import (
    WikiPage,
    WikiSection,
    WikiParagraph,
    WikiTitle,
    WikiAnchor,
    WikiInfoBoxProperty,
    WikiInfoBoxMapped,
    WikiCategory,
)

try:
    import rdflib
except ImportError as e:
    raise ImportError(
        create_import_error_msg(
            "rbflib", "wikipedia", "DBpedia dataset supports"
        )
    ) from e


__all__ = [
    "DBpediaWikiReader",
    "WikiPackReader",
    "WikiArticleWriter",
    "WikiAnchorReader",
    "WikiStructReader",
    "WikiPropertyReader",
    "WikiInfoBoxReader",
]


class DBpediaWikiReader(PackReader):
    """
    This reader reads in the Wikipedia full text articles from a DBpedia
    full text dump, which is the `NIF Context` dataset from here:
    https://wiki.dbpedia.org/downloads-2016-10#p10608-2 .
    """

    def __init__(
        self,
    ):
        super().__init__()
        self.__redirects: Dict[str, str] = {}

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)
        if self.resources.contains("redirects"):
            self.__redirects = self.resources.get("redirects")
            logging.info("%d redirects loaded.", len(self.__redirects))
        else:
            raise ResourceError("Redirects not provided from resources.")

    def _collect(  # type: ignore
        self, nif_context: str
    ) -> Iterator[Dict[str, str]]:
        str_data: Dict[str, str] = {}

        for context_statements in NIFParser(nif_context):
            for s, v, o, c in context_statements:
                nif_type = get_resource_attribute(s, "nif")
                print_progress(f"Collecting DBpedia resource: [{c.identifier}]")

                fragment = get_resource_fragment(v)
                if (
                    nif_type
                    and nif_type == "context"
                    and fragment is not None
                    and fragment == "isString"
                ):
                    str_data["text"] = o.toPython()
                    doc_name: Optional[str] = get_resource_name(s)
                    old_id: Optional[str] = get_resource_attribute(
                        c.identifier, "oldid"
                    )
                    if doc_name is not None and old_id is not None:
                        str_data["doc_name"] = doc_name
                        str_data["oldid"] = old_id
                        yield str_data

    def _parse_pack(self, doc_data: Dict[str, str]) -> Iterator[DataPack]:
        pack = DataPack()
        doc_name: str = doc_data["doc_name"]
        if doc_name in self.__redirects:
            doc_name = self.__redirects[doc_name]

        full_text: str = doc_data["text"]

        pack.set_text(full_text)
        page = WikiPage(pack, 0, len(full_text))
        page.page_id = doc_data["oldid"]
        page.page_name = doc_name
        pack.pack_name = doc_name
        yield pack


def read_index(pack_index_path: str) -> Dict[str, str]:
    """
    Reads an index from the page name to the path of the stored pack.

    Args:
        pack_index_path: The path of this index file. The file should be a tab
        separated file.

    Returns:
        A dictionary that maps from the page name to the full path.
    """
    page_idx: Dict[str, str] = {}

    logging.info("Reading pack index from %s", pack_index_path)

    with open(pack_index_path) as idx:
        for page_name, page_path in csv.reader(idx, delimiter="\t"):
            page_idx[page_name] = page_path
    return page_idx


class WikiPackReader(PackReader):
    """
    This reader reads information from an NIF graph, and find out the
    corresponding data pack stored on disk. The output from this reader
    are these data packs plus the additional NIF information.

    The function `add_wiki_info` is to be implemented to handle how the
    NIF statements are added to the data pack.
    """

    def __init__(self):
        super().__init__()
        self._pack_index: Dict[str, str] = {}
        self._pack_dir: str = ""
        self._redirects: Dict[str, str] = {}
        self._resume_index: Dict[str, str] = {}

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)

        # A mapping from the name of the page to the path on th disk.
        self._pack_index = read_index(configs.pack_index)
        self._pack_dir = configs.pack_dir

        if self.configs.resume_index:
            self._resume_index = read_index(configs.resume_index)
            print_progress(
                f"Loaded {len(self._resume_index)} existing " f"files.", "\n"
            )

        if self.resources.contains("redirects"):
            self._redirects = self.resources.get("redirects")
            print_progress(f"{len(self._redirects)} redirects loaded.", "\n")
        else:
            raise ResourceError("Redirects not provided from resources.")

    def add_wiki_info(self, pack: DataPack, statements: List[state_type]):
        raise NotImplementedError

    def _collect(  # type: ignore
        self, nif_path: str
    ) -> Iterator[Tuple[str, Dict[str, List[state_type]]]]:
        skipped = 0
        for _, statements in ContextGroupedNIFReader(nif_path):
            name = get_resource_name(statements[0][0])
            if name is not None:
                if name not in self._resume_index:
                    yield name, statements
                else:
                    skipped += 1
                    print_progress(
                        f"Skipped {skipped} documents", terminal_only=True
                    )

    def _parse_pack(
        self, collection: Tuple[str, List[state_type]]
    ) -> Iterator[DataPack]:
        resource_name, statements = collection
        if resource_name in self._redirects:
            resource_name = self._redirects[resource_name]

        if resource_name in self._pack_index:
            print_progress(
                f"Handling resource [{resource_name}] in {self.component_name}"
            )
            pack_path = os.path.join(
                self._pack_dir, self._pack_index[resource_name]
            )

            if os.path.exists(pack_path):
                pack: DataPack = DataPack.deserialize(
                    pack_path,
                    self.configs.serialize_method,
                    self.configs.zip_pack,
                )
                self.add_wiki_info(pack, statements)
                yield pack
        else:
            logging.info("Resource %s pack not found.", resource_name)

    @classmethod
    def default_configs(cls):
        """
        This defines a basic config structure for the reader.

        Here:
          - pack_dir: the directory that contains all the serialized packs.
          - pack_index: the file name under the pack directory that points to
            the index from the name to the actual pack path.

        """
        return {
            "pack_index": "article.idx",
            "pack_dir": ".",
            "resume_index": None,
        }


class WikiArticleWriter(PackWriter):
    """
    This is a pack writer that writes out the Wikipedia articles to disk. It
    has two special behaviors:
      1. If the `input_index_file` file is provided via the configuration and it
       exists, the file will be used to determine the path of writing the
       data packs. This will also activate the overwrite mode.
      2. If the `input_index_file` file is not provided,
         2a. The packs are organized into directories. Each directory contains
         at most 2000 documents.
         2b. the overwrite mode will not be activated
      3. If the `output_index_file` is provided, an index file with the provided
      name/path will be created, its content will be a mapping from the article
      name to the article path.

    There are two general use cases:
    1. If the writer is used to write a new directory of data, simply provide
    the `output_index_file`
    2. If the writer is used to add content/overwriting to an existing
    directory, it is suggested to use the index file of the original
    directory as the `input_index_file`, and the `output_index_file` can be
    used to store the information for this new writing process if desired.
    """

    _input_index_file: TextIO
    _output_index_file: TextIO

    # It is difficult to get the type of the csv writer
    # https://stackoverflow.com/questions
    # /51264355/how-to-type-annotate-object-returned-by-csv-writer
    _csv_writer: Any

    def __init__(self):
        super().__init__()
        self.article_count: int = 0
        self.__use_existing_index: bool = False
        self._article_index = {}

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)
        self.article_count = 0

        if self.configs.use_input_index and self.configs.input_index_file:
            # Load input index.
            input_index_path = self.configs.input_index_file
            self._article_index = {}
            if os.path.exists(input_index_path):
                self._input_index_file = open(input_index_path)
                with open(input_index_path) as f:
                    for line in f:
                        article_name, sub_path = line.strip().split()
                        self._article_index[article_name] = sub_path
                self.__use_existing_index = True
                self.configs.overwrite = True
                logging.info(
                    "Wikipedia writer is setup with existing index "
                    "file. The output will be written following the input  "
                    "index path and overwritten is enabled."
                )
            else:
                raise FileNotFoundError(
                    f"Cannot find provided index file {input_index_path}"
                )
        else:
            self.__use_existing_index = False

        output_index_path = os.path.join(
            self.configs.output_dir, self.configs.output_index_file
        )
        self._output_index_file = (
            open(output_index_path, "a")
            if self.configs.append_to_index
            else open(output_index_path, "w")
        )
        self._csv_writer = csv.writer(self._output_index_file, delimiter="\t")

    def sub_output_path(self, pack: DataPack) -> Optional[str]:
        if self.__use_existing_index:
            if pack.pack_name in self._article_index:
                # Since datasets are built separated, there might be cases
                # where the article referred later is not in the original
                # parsed dataset, so we need to check if they exist.

                # We could replace the suffix based on writing config.
                return (
                    self._article_index[pack.pack_name].split(".")[0]
                    + self._suffix
                )
            else:
                return None
        else:
            # Organize the data by IO ordering instead.
            sub_dir = str(int(self.article_count / 2000)).zfill(5)
            pid = pack.get_single(WikiPage).page_id  # type: ignore
            doc_name = f"doc_{self.article_count}" if pid is None else pid
            return os.path.join(sub_dir, doc_name) + self._suffix

    def _process(self, input_pack: DataPack):
        """
        In additional writing the data pack, we also write the index under
        the condition, to store the document id to the relative storage of this
        DataPack. This can be used as a simple index to retrieve the relevant
        file, which can enable faster lookup.

        Args:
            input_pack: The DataPack that contains the Wikipedia information.

        Returns:
        """
        super()._process(input_pack)

        # Write the output index files.
        out_path = self.sub_output_path(input_pack)
        self._csv_writer.writerow([input_pack.pack_name, out_path])

        self.article_count += 1
        if self.article_count % 1000 == 0:
            logging.info(
                "Written %s to %s", self.article_count, self.configs.output_dir
            )

    def finish(self, _: Resources):
        if self.configs.use_input_index and self.configs.input_index_file:
            self._input_index_file.close()
        self._output_index_file.close()

    @classmethod
    def default_configs(cls):
        """
        This defines a basic config structure for the reader.

        The additional parameters to provide:
          - use_input_index (bool): whether to use the input index file to find
              data.

          - input_index_file (str): the path providing the index from the
              wikipedia article name to the relative paths that stores these
              files.
              This file will only be used if the `use_input_index` and
              `overwrite` are both set to true, and the data path will be
              used to write the results (which means the existing files will be
              overwritten).

          - output_index_file (str): if provided, will write out the index from
              file name to the packs. This path and the relative paths are all
              relative names are relative to the `output_dir`.

          - append_to_index (bool): if provided, will append to the
             `output_index_file` instead of creating a new one.

        Returns: The default configuration of this writer.
        """
        return {
            "use_input_index": False,
            "input_index_file": None,
            "output_index_file": "article.idx",
            "append_to_index": False,
        }


class WikiStructReader(WikiPackReader):
    """
    This reader extends the WikiPackReader and add the Wikipedia Structure
    information from https://wiki.dbpedia.org/downloads-2016-10#p10608-2
    """

    def add_wiki_info(self, pack: DataPack, statements: List):
        for nif_range, rel, struct_type in statements:
            r = get_resource_fragment(rel)
            if r is not None and r == "type":
                range_ = get_resource_attribute(nif_range, "char")
                if range_ is None:
                    continue

                begin, end = [int(d) for d in range_.split(",")]

                if end > len(pack.text):
                    # Some nif dataset are off by a bit, mostly when there
                    # are new line characters, we cannot correct them.
                    # but we need to make sure they don't go longer than
                    # the text.
                    logging.info(
                        "NIF Structure end is %d by %s, "
                        "clipped to fit with the text.",
                        end,
                        nif_range,
                    )
                    end = len(pack.text)

                if end <= begin:
                    logging.info(
                        "Provided struct [%d:%d] is invalid.", begin, end
                    )
                    continue

                struct_ = get_resource_fragment(struct_type)

                if struct_ is not None:
                    if struct_ == "Section":
                        WikiSection(pack, begin, end)
                    elif struct_ == "Paragraph":
                        WikiParagraph(pack, begin, end)
                    elif struct_ == "Title":
                        WikiTitle(pack, begin, end)
                    else:
                        logging.warning("Unknown struct type: %s", struct_type)


class WikiAnchorReader(WikiPackReader):
    """
    This reader extends the WikiPackReader and add the Wikipedia anchors
    from https://wiki.dbpedia.org/downloads-2016-10#p10608-2
    """

    def add_wiki_info(self, pack: DataPack, statements: List):
        link_grouped: DefaultDict[
            str, Dict[str, rdflib.term.Node]
        ] = defaultdict(dict)
        for nif_range, rel, info in statements:
            range_ = get_resource_attribute(nif_range, "char")
            r = get_resource_fragment(rel)
            if range_ is not None and r is not None:
                link_grouped[range_][r] = info

        for range_, link_infos in link_grouped.items():
            begin, end = [int(d) for d in range_.split(",")]

            if end > len(pack.text):
                # Some nif dataset are off by a bit, mostly when there are
                # new line characters, we cannot correct them.
                # but we need to make sure they don't go longer than the
                # text.
                logging.info(
                    "Provided anchor end is %d, "
                    "clipped to fit with the text.",
                    end,
                )
                end = len(pack.text)

            if end <= begin:
                logging.info("Provided anchor [%d:%d is invalid.]", begin, end)
                continue

            for info_key, info_value in link_infos.items():
                info_value = str(info_value)
                if info_key == "type":
                    anchor_type = get_resource_fragment(info_value)
                    if (
                        not anchor_type == "Phrase"
                        and not anchor_type == "Word"
                    ):
                        logging.warning("Unknown anchor type: %s", info_value)
                if info_key == "taIdentRef":
                    target_page_name = get_resource_name(info_value)
                    if (
                        target_page_name is not None
                        and target_page_name in self._redirects
                    ):
                        target_page_name = self._redirects[target_page_name]

                    if target_page_name is not None:
                        # Only create anchor with proper link.
                        anchor = WikiAnchor(pack, begin, end)
                        anchor.target_page_name = target_page_name
                        # If it is an DBpedia resource, the domain will be
                        # truncated, otherwise it will stay the same, meaning
                        # it is an external link.
                        anchor.is_external = target_page_name == str(info_value)


class WikiPropertyReader(WikiPackReader):
    """
    This reader extends the WikiPackReader and add the Wikipedia raw info
    boxes (also known as properties) from
    https://wiki.dbpedia.org/downloads-2016-10#p10608-2
    """

    def add_wiki_info(self, pack: DataPack, statements: List):
        for _, v, o in statements:
            slot_name = v.toPython()
            slot_value = get_resource_name(o)
            if slot_value is not None:
                info_box = WikiInfoBoxProperty(pack)
                info_box.key = slot_name
                info_box.value = slot_value


class WikiInfoBoxReader(WikiPackReader):
    """
    This reader extends the WikiPackReader and add the Wikipedia cleaned
    info boxes from https://wiki.dbpedia.org/downloads-2016-10#p10608-2
    """

    def add_wiki_info(self, pack: DataPack, info_box_statements: List):
        for _, v, o in info_box_statements:
            name = get_resource_name(o)
            if name is not None:
                info_box = WikiInfoBoxMapped(pack)
                info_box.key = v.toPython()
                info_box.value = name


class WikiCategoryReader(WikiPackReader):
    """
    Read the dbpedia category file to add category information.
    """

    def add_wiki_info(self, pack: DataPack, statements: List[state_type]):
        for _, _, o in statements:
            resource_name = get_resource_name(o)
            if resource_name is not None:
                wc = WikiCategory(pack)
                wc.values.append(resource_name)
                pack.add_entry(wc)



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Base reader type to be inherited by all readers.
"""
import logging
import os
from time import time
from abc import abstractmethod, ABC
from pathlib import Path
from typing import Any, Iterator, Optional, Union, List, Dict, Set

from forte.common.configuration import Config
from forte.common.exception import ProcessExecutionException
from forte.common.resources import Resources
from forte.data.base_pack import PackType
from forte.data.data_pack import DataPack
from forte.data.multi_pack import MultiPack
from forte.data.types import ReplaceOperationsType
from forte.pipeline_component import PipelineComponent
from forte.utils.utils import get_full_module_name

__all__ = [
    "BaseReader",
    "PackReader",
    "MultiPackReader",
]

logger = logging.getLogger(__name__)


class BaseReader(PipelineComponent[PackType], ABC):
    r"""The basic data reader class. To be inherited by all data readers.

    Args:
        from_cache: Decide whether to read from cache
            if cache file exists. By default (``False``), the reader will
            only read from the original file and use the cache file path
            for caching, it will not read from the ``cache_directory``.
            If ``True``, the reader will try to read a datapack from the
            caching file.
        cache_directory: The base directory to place the
            path of the caching files. Each collection is contained in one
            cached file, under this directory. The cached location for each
            collection is computed by
            :meth:`~forte.data.base_reader.BaseReader._cache_key_function`.

            .. note::

                A collection is the data returned by
                :meth:`~forte.data.base_reader.BaseReader._collect`.

        append_to_cache: Decide whether to append write
            if cache file already exists.  By default (``False``), we
            will overwrite the existing caching file. If ``True``, we will
            cache the datapack append to end of the caching file.
    """

    def __init__(
        self,
        from_cache: bool = False,
        cache_directory: Optional[str] = None,
        append_to_cache: bool = False,
        cache_in_memory: bool = False,
    ):
        super().__init__()
        self.from_cache = from_cache
        self._cache_directory = cache_directory
        self.component_name = get_full_module_name(self)
        self.append_to_cache = append_to_cache
        self._cache_in_memory = cache_in_memory
        self._cache_ready: bool = False
        self._data_packs: List[PackType] = []

        # needed for time profiling of reader
        self._enable_profiling: bool = False
        self._start_time: float = 0.0
        self.time_profile: float = 0.0

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)

        # Clear memory cache
        self._cache_ready = False
        del self._data_packs[:]

    @classmethod
    def default_configs(cls):
        r"""Returns a `dict` of configurations of the reader with default
        values. Used to replace the missing values of input `configs`
        during pipeline construction.

        Here:

          - zip_pack (bool): whether to zip the results. The default value is
            False.

          - serialize_method: The method used to serialize the data. Current
            available options are `jsonpickle` and `pickle`. Default is
            `jsonpickle`.

        """
        return {"zip_pack": False, "serialize_method": "jsonpickle"}

    @staticmethod
    def pack_type():
        raise NotImplementedError

    @abstractmethod
    def _collect(self, *args: Any, **kwargs: Any) -> Iterator[Any]:
        r"""Returns an iterator of data objects, and each individual object
        should contain sufficient information needed to construct or locate
        a data pack in cache.

        For example: `data_source` can be a ``kwarg`` which is the path to a
        file that a reader can take to read and parse a file.

        Args:
            args: Specify the data source.
            kwargs: Specify the data source.

        Returns: Iterator of collections that are sufficient to create one pack.
        """
        raise NotImplementedError

    def parse_pack(self, collection: Any) -> Iterator[PackType]:
        r"""Calls :meth:`_parse_pack` to create packs from the collection.
        This internally setup the component meta data. Users should implement
        the :meth:`_parse_pack` method.
        """
        if collection is None:
            raise ProcessExecutionException(
                "Got None collection, cannot parse as data pack."
            )

        for p in self._parse_pack(collection):
            p.add_all_remaining_entries(self.name)
            yield p

    @abstractmethod
    def _parse_pack(self, collection: Any) -> Iterator[PackType]:
        r"""Returns an iterator of Packs parsed from a collection. Readers
        should implement this class to populate the class input.

        Args:
            collection: Object that can be parsed into a Pack.

        Returns: Iterator of Packs.
        """
        raise NotImplementedError

    def _cache_key_function(self, collection: Any) -> Optional[str]:
        # pylint: disable=unused-argument
        r"""Computes the cache key based on the type of data.

        Args:
            collection: Any object that provides information to identify the
                name and location of the cache file
        """
        return None

    def text_replace_operation(self, text: str) -> ReplaceOperationsType:
        # pylint: disable=unused-argument
        r"""Given the possibly noisy text, compute and return the
        replacement operations in the form of a list of (span, str)
        pairs, where the content in the span will be replaced by the
        corresponding str.

        Args:
            text: The original data text to be cleaned.

        Returns (List[Tuple[Tuple[int, int], str]]):
            the replacement operations.
        """
        return []

    def _get_cache_location(self, collection: Any) -> str:
        r"""Gets the path to the cache file for a collection.

        Args:
            collection: information to compute cache key.

        Returns (Path): file path to the cache file for a Pack.
        """
        # pylint: disable=assignment-from-none
        file_path = self._cache_key_function(collection)
        if file_path is None:
            raise ProcessExecutionException(
                "Cache key is None. You probably set `from_cache` to true but "
                "fail to implement the _cache_key_function"
            )

        return os.path.join(str(self._cache_directory), file_path)

    def _lazy_iter(self, *args, **kwargs):
        for collection in self._collect(*args, **kwargs):
            if self.from_cache:
                for pack in self.read_from_cache(
                    self._get_cache_location(collection)
                ):
                    pack.add_all_remaining_entries()
                    yield pack
            else:
                not_first = False
                for pack in self.parse_pack(collection):
                    # write to the cache if _cache_directory specified
                    if self._cache_directory is not None:
                        self.cache_data(collection, pack, not_first)

                    if not isinstance(pack, self.pack_type()):
                        raise ValueError(
                            f"No Pack object read from the given "
                            f"collection {collection}, returned {type(pack)}."
                        )

                    not_first = True
                    pack.add_all_remaining_entries()
                    yield pack

    def set_profiling(self, enable_profiling: bool = True):
        r"""Set profiling option.

        Args:
            enable_profiling: A boolean of whether to enable profiling
                for the reader or not (the default is True).
        """
        self._enable_profiling = enable_profiling

    def timer_yield(self, pack: PackType):
        r"""Wrapper generator for time profiling. Insert timers around
        'yield' to support time profiling for reader.

        Args:
            pack: DataPack passed from self.iter()
        """
        # Aggregate time cost
        if self._enable_profiling:
            self.time_profile += time() - self._start_time

        yield pack

        # Start timer
        if self._enable_profiling:
            self._start_time = time()

    def iter(self, *args, **kwargs) -> Iterator[PackType]:
        # pylint: disable=protected-access
        r"""An iterator over the entire dataset, giving all Packs processed
        as list or Iterator depending on `lazy`, giving all the Packs read
        from the data source(s). If not reading from cache, should call
        ``collect``.

        Args:
            args: One or more input data sources, for example, most
                DataPack readers accept `data_source` as file/folder path.
            kwargs: Iterator of DataPacks.
        """

        # Start timer
        if self._enable_profiling:
            self._start_time = time()

        if self._cache_in_memory and self._cache_ready:
            # Read from memory
            for pack in self._data_packs:
                if self._check_type_consistency:
                    if hasattr(pack._meta, "record"):
                        self.record(pack._meta.record)
                yield from self.timer_yield(pack)
        else:
            # Read via parsing dataset
            for pack in self._lazy_iter(*args, **kwargs):
                if self._check_type_consistency:
                    if hasattr(pack._meta, "record"):
                        self.record(pack._meta.record)
                if self._cache_in_memory:
                    self._data_packs.append(pack)

                yield from self.timer_yield(pack)

        self._cache_ready = True

    def record(self, record_meta: Dict[str, Set[str]]):
        r"""Modify the pack meta record field of the reader's output. The
        key of the record should be the entry type and values should
        be attributes of the entry type. All the information would be used
        for consistency checking purpose if the pipeline is initialized with
        `enforce_consistency=True`.

        Args:
            record_meta: the field in the datapack for type record that need to
                fill in for consistency checking.
        """
        pass

    def cache_data(self, collection: Any, pack: PackType, append: bool):
        r"""Specify the path to the cache directory.

        After you call this method, the dataset reader will use its
        ``cache_directory`` to store a cache of
        :class:`~forte.data.base_pack.BasePack` read
        from every document passed to read, serialized as one
        string-formatted :class:`~forte.data.base_pack.BasePack`. If the cache file for a given
        ``file_path`` exists, we read the :class:`~forte.data.base_pack.BasePack` from the cache.
        If the cache file does not exist, we will `create` it on our first
        pass through the data.

        Args:
            collection: The collection is a piece of data from the
                :meth:`_collect` function, to be read to produce DataPack(s).
                During caching, a cache key is computed based on the data in
                this collection.
            pack: The data pack to be cached.
            append: Whether to allow appending to the cache.
        """
        if not self._cache_directory:
            raise ValueError("Can not cache without a cache_directory!")

        os.makedirs(self._cache_directory, exist_ok=True)

        cache_filename = os.path.join(
            self._cache_directory, self._get_cache_location(collection)
        )

        logger.info("Caching pack to %s", cache_filename)
        if append:
            with open(
                cache_filename,
                "a",
                encoding="utf-8",
            ) as cache:
                cache.write(pack.to_string() + "\n")
        else:
            with open(
                cache_filename,
                "w",
                encoding="utf-8",
            ) as cache:
                cache.write(pack.to_string() + "\n")

    def read_from_cache(
        self, cache_filename: Union[Path, str]
    ) -> Iterator[PackType]:
        r"""Reads one or more Packs from ``cache_filename``, and yields Pack(s)
        from the cache file.

        Args:
            cache_filename: Path to the cache file.

        Returns:
            List of cached data packs.
        """
        logger.info("reading from cache file %s", cache_filename)
        with open(cache_filename, "r", encoding="utf-8") as cache_file:
            for line in cache_file:
                pack = DataPack.from_string(line.strip())
                if not isinstance(pack, self.pack_type()):
                    raise TypeError(
                        f"Pack deserialized from {cache_filename} "
                        f"is {type(pack)}, but expect {self.pack_type()}"
                    )
                yield pack

    def finish(self, resource: Resources):
        pass

    def set_text(self, pack: DataPack, text: str):
        r"""Assign the text value to the
        :class:`~forte.data.data_pack.DataPack`. This function will
        pass the ``text_replace_operation`` to the
        :class:`~forte.data.data_pack.DataPack` to conduct
        the pre-processing step.

        Args:
            pack: The :class:`~forte.data.data_pack.DataPack` to assign value for.
            text: The original text to be recorded in this dataset.
        """
        pack.set_text(text, replace_func=self.text_replace_operation)


class PackReader(BaseReader[DataPack], ABC):
    r"""A Pack Reader reads data into :class:`~forte.data.data_pack.DataPack`."""

    @staticmethod
    def pack_type():
        return DataPack


class MultiPackReader(BaseReader[MultiPack], ABC):
    r"""The basic :class:`~forte.data.multi_pack.MultiPack` data reader class.
    To be inherited by all
    data readers which return :class:`~forte.data.multi_pack.MultiPack`.
    """

    @staticmethod
    def pack_type():
        return MultiPack



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import random
from abc import abstractmethod
from typing import Any, Dict, Union
from forte.common.configurable import Configurable

from forte.common.configuration import Config


__all__ = [
    "Sampler",
    "UniformSampler",
    "UnigramSampler",
]


class Sampler(Configurable):
    r"""
    An abstract sampler class.
    """

    def __init__(self, configs: Union[Config, Dict[str, Any]]):
        self.configs: Config = self.make_configs(configs)
        random.seed()

    @abstractmethod
    def sample(self) -> str:
        raise NotImplementedError


class UniformSampler(Sampler):
    r"""
    A sampler that samples a word from a uniform distribution.

    Config Values:

        - sampler_data:
          a list of words that this sampler uniformly samples from.
    """

    def __init__(self, configs: Union[Config, Dict[str, Any]]):
        super().__init__(configs)
        self.word_list = self.configs["sampler_data"]

    def sample(self) -> str:
        word: str = random.choice(self.word_list)
        return word

    @classmethod
    def default_configs(cls):
        return {"sampler_data": [], "@no_typecheck": "sampler_data"}


class UnigramSampler(Sampler):
    r"""
    A sampler that samples a word from a unigram distribution.

    Config Values:

        - sampler_data: (dict)
          The key is a word, the value is the word count or a probability.
          This sampler samples from this word distribution.
          Example:

            .. code-block:: python

                'sampler_data': {
                        "apple": 1,
                        "banana": 2,
                        "orange": 3
                }

    """

    def __init__(self, configs: Union[Config, Dict[str, Any]]):
        super().__init__(configs)
        self.unigram = self.configs["sampler_data"].__dict__["_hparams"]

    def sample(self) -> str:
        word: str = random.choices(
            list(self.unigram.keys()), list(self.unigram.values())
        )[0]
        return word

    @classmethod
    def default_configs(cls):
        return {"sampler_data": {}, "@no_typecheck": "sampler_data"}



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Processors that augment the data. The processor will call
replacement ops to generate texts similar to those in the input pack
and create a new pack with them.
"""
from abc import ABC
from bisect import bisect_right, bisect_left
from collections import defaultdict
from copy import deepcopy
from typing import List, Tuple, Dict, DefaultDict, Set, Union, cast, Iterable

from sortedcontainers import SortedList, SortedDict

from forte.common.configuration import Config
from forte.common.resources import Resources
from forte.data.data_pack import DataPack
from forte.data.multi_pack import MultiPack
from forte.data.ontology.core import Entry, BaseLink
from forte.data.ontology.top import (
    Annotation,
    MultiPackLink,
    Link,
    MultiPackGroup,
    Group,
)
from forte.data.span import Span
from forte.processors.base import MultiPackProcessor
from forte.processors.data_augment.algorithms.text_replacement_op import (
    TextReplacementOp,
)
from forte.utils.utils import get_class, create_class_with_kwargs

__all__ = ["BaseDataAugmentProcessor", "ReplacementDataAugmentProcessor"]


def modify_index(
    index: int,
    # Both of the following spans should be SortedList.
    # Use List to avoid typing errors.
    old_spans: List[Span],
    new_spans: List[Span],
    is_begin: bool,
    is_inclusive: bool,
) -> int:
    r"""
    A helper function to map an index before replacement

    to the index after replacement.
    An index is the character offset in the data pack.
    The `old_spans` are the inputs of replacement, and the new_spans
    are the outputs. Each of the span has start and end index.
    The `old_spans` and `new_spans` are anchors for the mapping,
    because we depend on them to determine the position change of the
    index.

    Given an index, the function will find its the nearest
    among the old spans before the index, and calculate the difference
    between the position of the old span and its corresponding new span.
    The position change is then applied to the input index. An updated
    index is then calculated and returned.

    An inserted span might be included as a part of another span.
    For example, given a sentence "I love NLP.", if we insert a
    token "Yeah" at the beginning of the sentence(index=0), the Sentence
    should include the new Token, i.e., the Sentence will have a
    start index equals to 0. In this case, the parameter is_inclusive
    should be True. However, for another Token "I", it should not include
    the new token, so its start index will be larger than 0.
    The parameter in_inclusive should be False.

    The input index could be the start or end index of a span, i.e., the
    left or right boundary of the span. If there is an insertion in the span,
    we should treat the two boundaries in different ways. For example,
    we have a paragraph with two sentences "I love NLP! You love NLP too."
    If we append another "!" to the end of the first sentence, when modifying
    the end index of the first Sentence, it should be pushed right to include
    the extra exclamation. In this case, the is_begin is False. However, if
    we prepend an "And" to the second sentence, when modifying the start index
    of the second Sentence, it should be pushed left to include the new Token.
    In this case, the `is_begin` is True.
    Args:
        index (int): The index to map.
        old_spans (SortedList): The spans before replacement. It should be
            a sorted list in ascending order.
        new_spans (SortedList): The spans after replacement. It should be
            a sorted list in ascending order.
        is_begin (bool): True if the input index is the start index of a span.
        is_inclusive (bool): True if the span constructed by the aligned
            index should include inserted spans.
    Returns:
        The aligned index.
    If the old spans are [0, 1], [2, 3], [4, 6],
    the new spans are [0, 4], [5, 7], [8, 11],
    the input index is 3, and there are no insertions,
    the algorithm will first locate the last span with
    a begin index less or equal than the target index,
    ([2,3]), and find the corresponding span in new spans([5,7]).
    Then we calculate the delta index(7-3=4) and update our
    input index(3+4=7). The output then is 7.

    Note that when the input index locates inside the old spans,
    instead of on the boundary of the spans, we compute the return
    index so that it maintains the same offset to the begin of the
    span it belongs to. In the above example, if we change the input
    index from 3 to 5, the output will become 9, because we locates
    the input index in the third span [4, 6] and use the same offset
    5-4=1 to calculate the output 8+1=9.

    When insertion is considered, there will be spans
    with the same begin index, for example,
    [0, 1], [1, 1], [1, 2]. The span [1, 1] indicates an insertion
    at index 1, because the insertion can be considered as a
    replacement of an empty input span, with a length of 0.
    The output will be affected by whether to include the inserted
    span(is_inclusive), and whether the input index is a begin or
    end index of its span(is_begin).

    If the old spans are [0, 1], [1, 1], [1, 2],
    the new spans are [0, 2], [2, 4], [4, 5],
    the input index is 1, the output will be 2 if both
    is_inclusive and is_begin are True,
    because the inserted [1, 1] should be included in the span.
    If the `is_inclusive=True` but `is_begin=False`, the output will be
    4 because the index is an end index of the span.
    """

    # Get the max index for binary search.
    max_index: int = old_spans[-1].end + 1
    max_index = max(max_index, index)

    # This is the last span that has a start index less than
    # the input index. The position change of this span determines
    # the modification we will apply to the input index.
    last_span_ind: int = bisect_right(old_spans, Span(index, max_index)) - 1

    # If there is an inserted span, it will always be the first of
    # those spans with the same begin index. For example, given spans
    # [1, 1], [1, 2], The inserted span [1, 1] will be in the front of
    # replaced span [1, 2], because it has the smallest end index.
    if last_span_ind >= 0:
        if is_inclusive:
            if is_begin:
                # When inclusive, move the begin index
                # to the left to include the inserted span.
                if (
                    last_span_ind > 0
                    and old_spans[last_span_ind - 1].begin == index
                ):
                    # Old spans: [0, 1], [1, 1], [1, 3]
                    # Target index: 1
                    # Change last_span_index from 2 to 1
                    # to include the [1, 1] span.
                    last_span_ind -= 1
                else:
                    # Old spans: [0, 1], [1, 1], [2, 3]
                    # Target index: 1
                    # last_span_index: 1
                    # No need to change.
                    pass

        else:
            if not is_begin:
                # When exclusive, move the end index
                # to the left to exclude the inserted span.
                if (
                    last_span_ind > 0
                    and old_spans[last_span_ind - 1].begin == index
                ):
                    # Old spans: [0, 1], [1, 1], [1, 3]
                    # Target index: 1
                    # Change last_span_index from 2 to 0
                    # to exclude the [1, 1] span.
                    last_span_ind -= 2
                elif (
                    old_spans[last_span_ind].begin == index
                    and old_spans[last_span_ind].end == index
                ):
                    # Old spans: [0, 1], [1, 1], [2, 3]
                    # Target index: 1
                    # Change last_span_index from 1 to 0
                    # to exclude the [1, 1] span.
                    last_span_ind -= 1

    if last_span_ind < 0:
        # There is no replacement before this index.
        return index
    # Find the nearest anchor point on the left of current index.
    # Start from the span's begin index.
    delta_index: int = (
        new_spans[last_span_ind].begin - old_spans[last_span_ind].begin
    )

    if (
        old_spans[last_span_ind].begin == old_spans[last_span_ind].end
        and old_spans[last_span_ind].begin == index
        and is_begin
        and is_inclusive
    ):
        return index + delta_index

    if old_spans[last_span_ind].end <= index:
        # Use the span's end index as anchor, if possible.
        delta_index = (
            new_spans[last_span_ind].end - old_spans[last_span_ind].end
        )
    return index + delta_index


class BaseDataAugmentProcessor(MultiPackProcessor, ABC):
    r"""
    The base class of processors that augment data.
    This processor instantiates replacement ops where specific
    data augmentation algorithms are implemented. The replacement ops
    will run the algorithms and the processor will create Forte
    data structures based on the augmented inputs.
    """


class ReplacementDataAugmentProcessor(BaseDataAugmentProcessor):
    r"""
    Most of the Data Augmentation(DA) methods can be
    considered as replacement-based methods with different
    levels: character, word, sentence or document.
    """

    def __init__(self):
        super().__init__()
        # :attr:`_replaced_annos`: {datapack id: SortedList[span, new text]}
        # It records the spans replaced by new texts.
        # It is a map from datapack id to a list of tuples
        # (span, new text) inserted by :func:`replace`.
        # The new text will be used for building new data pack.
        self._replaced_annos: DefaultDict[
            int, SortedList[Tuple[Span, str]]
        ] = defaultdict(lambda: SortedList([], key=lambda x: x[0]))

        # :attr:`_inserted_annos_pos_len`: {datapack id: Dict{position: length}}
        # It records the position and length of inserted spans,
        # mapping from datapack id to a dictionary (position -> length)
        # inserted by :func:`insert`.
        # The position is the index in the original datapack
        # of insertion, and the length is the length of the inserted string.
        self._inserted_annos_pos_len: DefaultDict[
            int, SortedDict[int, int]
        ] = defaultdict(lambda: SortedDict({}))

        # :attr:`_deleted_annos_id`: {datapack id: Set[annotation tid]}
        # It records the deleted span ids, mapping from datapack id
        # to a set of annotation tids appended by :func:`delete`.
        self._deleted_annos_id: DefaultDict[int, Set[int]] = defaultdict(set)

        # :attr:`_data_pack_map`: {orig pack id: new pack id}
        # It maintains a mapping from the pack id
        # of the original pack to the pack id of augmented pack.
        # It is used when copying the MultiPackLink and MultiPackGroup.
        self._data_pack_map: Dict[int, int] = {}

        # :attr:`_entry_maps`: {datapack id: Dict{orig tid, new tid}}
        # It is a map for tracking the annotation ids
        # before and after the auto align. It maps the
        # original annotation tid to the new annotation tid.
        # It is used when copying the Link/Group/MultiPackLink/MultiPackGroup.
        self._entry_maps: Dict[int, Dict[int, int]] = {}
        self._other_entry_policy: Dict[str, str] = {}

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)
        self._other_entry_policy = self.configs["other_entry_policy"]

    def _overlap_with_existing(self, pid: int, begin: int, end: int) -> bool:
        r"""
        This function will check whether the new span
        has an overlap with any existing spans.
        Args:
            pid: Datapack Id.
            begin: The span begin index.
            end: The span end index.
        Returns:
            True if the input span overlaps with
            any existing spans, False otherwise.
        """
        if len(self._replaced_annos[pid]) == 0:
            return False
        ind: int = (
            bisect_left(self._replaced_annos[pid], (Span(begin, begin), "")) - 1
        )

        if ind < 0:
            ind += 1

        while ind < len(self._replaced_annos[pid]):
            span: Span = self._replaced_annos[pid][ind][0]
            if not (span.begin >= end or span.end <= begin):
                return True
            if span.begin > end:
                break
            ind += 1

        return False

    def _replace(
        self, replacement_op: TextReplacementOp, input_anno: Annotation
    ) -> bool:
        r"""
        This is a wrapper function to call the replacement op. After
        getting the augmented text, it will register the input & output
        for later batch process of building the new data pack.
        It will ignore the input if it has an overlap with the already
        augmented spans.
        Args:
            replacement_op: The class for data augmentation algorithm.
            input_anno: The entry to be replaced.
        Returns:
            A bool value. True if the replacement happened, False otherwise.
        """
        # Ignore the new annotation if overlap.
        pid: int = input_anno.pack.pack_id
        if self._overlap_with_existing(pid, input_anno.begin, input_anno.end):
            return False
        replaced_text: str
        is_replace: bool
        is_replace, replaced_text = replacement_op.replace(input_anno)
        if is_replace:
            self._replaced_annos[pid].add((input_anno.span, replaced_text))
            return True
        return False

    def _insert(
        self, inserted_text: str, data_pack: DataPack, pos: int
    ) -> bool:
        r"""
        This is a wrapper function to insert a new annotation. After
        getting the inserted text, it will register the input & output
        for later batch process of building the new data pack.

        The insertion at each position can only occur once. If there
        is already an insertion at current position, it will abort the
        insertion and return False.
        Args:
            inserted_text: The text string to insert.
            data_pack: The datapack for insertion.
            pos: The position(index) of insertion.
        Returns:
            A bool value. True if the insertion happened, False otherwise.
        """
        pid: int = data_pack.pack_id
        if self._overlap_with_existing(pid, pos, pos):
            return False

        if pos not in self._inserted_annos_pos_len[pid]:
            self._replaced_annos[pid].add((Span(pos, pos), inserted_text))
            self._inserted_annos_pos_len[pid][pos] = len(inserted_text)
            return True
        return False

    def _delete(self, input_anno: Annotation) -> bool:
        r"""
        This is a wrapper function to delete an annotation.
        Args:
            input_anno: The annotation to remove.
        Returns:
            A bool value. True if the deletion happened, False otherwise.
        """
        pid: int = input_anno.pack.pack_id
        self._replaced_annos[pid].add((input_anno.span, ""))
        self._deleted_annos_id[pid].add(input_anno.tid)
        return True

    def _auto_align_annotations(
        self,
        data_pack: DataPack,
        replaced_annotations: SortedList,
    ) -> DataPack:
        r"""
        Function to replace some annotations with new strings.
        It will copy and update the text of datapack and
        auto-align the annotation spans.
        The links are also copied if its parent & child are
        both present in the new pack.
        The groups are copied if all its members are present
        in the new pack.
        Args:
            data_pack: The Datapack holding the replaced annotations.
            replaced_annotations: A SortedList of tuples(span, new string).
                The text and span of the annotations will be updated
                with the new string.
        Returns:
            A new data_pack holds the text after replacement. The annotations
            in the original data pack will be copied and auto-aligned as
            instructed by the "other_entry_policy" in the configuration.
            The links and groups will be copied if there members are copied.
        """
        if len(replaced_annotations) == 0:
            return deepcopy(data_pack)

        spans: List[Span] = [span for span, _ in replaced_annotations]
        replacement_strs: List[str] = [
            replacement_str for _, replacement_str in replaced_annotations
        ]

        # Get the new text for the new data pack.
        new_text: str = ""
        for i, span in enumerate(spans):
            new_span_str = replacement_strs[i]
            # First, get the gap text between last and this span.
            last_span_end: int = spans[i - 1].end if i > 0 else 0
            gap_text: str = data_pack.text[last_span_end : span.begin]
            new_text += gap_text
            # Then, append the replaced new text.
            new_text += new_span_str
        # Finally, append to new_text the text after the last span.
        new_text += data_pack.text[spans[-1].end :]

        # Get the span (begin, end) before and after replacement.
        new_spans: List[Span] = []

        # Bias is the delta between the beginning
        # indices before & after replacement.
        bias: int = 0
        for i, span in enumerate(spans):
            old_begin: int = spans[i].begin
            old_end: int = spans[i].end
            new_begin: int = old_begin + bias
            new_end = new_begin + len(replacement_strs[i])
            new_spans.append(Span(new_begin, new_end))
            bias = new_end - old_end

        new_pack: DataPack = DataPack()
        new_pack.set_text(new_text)

        entries_to_copy: List[str] = list(self._other_entry_policy.keys()) + [
            self.configs["augment_entry"]
        ]

        entry_map: Dict[int, int] = {}
        insert_ind: int = 0
        pid: int = data_pack.pack_id

        inserted_annos: List[Tuple[int, int]] = list(
            self._inserted_annos_pos_len[pid].items()
        )

        def _insert_new_span(
            entry_class: str,
            insert_ind: int,
            inserted_annos: List[Tuple[int, int]],
            new_pack: DataPack,
            spans: List[Span],
            new_spans: List[Span],
        ):
            """
            An internal helper function for insertion.
            Args:
                entry_class: The new annotation type to be created.
                insert_ind: The index to be insert.
                inserted_annos: The annotation span information to be inserted.
                new_pack: The new data pack to insert the annotation.
                spans: The original spans before replacement, should be
                  a sorted ascending list.
                new_spans: The original spans before replacement, should be
                  a sorted ascending list.
            Returns:
            """
            pos: int
            length: int
            pos, length = inserted_annos[insert_ind]
            insert_end: int = modify_index(
                pos,
                spans,
                new_spans,
                is_begin=False,
                # Include the inserted span itself.
                is_inclusive=True,
            )
            insert_begin: int = insert_end - length
            new_anno = create_class_with_kwargs(
                entry_class,
                {"pack": new_pack, "begin": insert_begin, "end": insert_end},
            )
            new_pack.add_entry(new_anno)

        # Iterate over all the original entries and modify their spans.
        for entry_to_copy in entries_to_copy:
            class_to_copy = get_class(entry_to_copy)
            if not issubclass(class_to_copy, Annotation):
                raise AttributeError(
                    f"The entry type to copy from [{entry_to_copy}] is not "
                    f"a sub-class of 'forte.data.ontology.top.Annotation'."
                )

            orig_annos: Iterable[Annotation] = data_pack.get(class_to_copy)
            for orig_anno in orig_annos:
                # Dealing with insertion/deletion only for augment_entry.
                if entry_to_copy == self.configs["augment_entry"]:
                    while (
                        insert_ind < len(inserted_annos)
                        and inserted_annos[insert_ind][0] <= orig_anno.begin
                    ):
                        # Preserve the order of the spans with merging sort.
                        # It is a 2-way merging from the inserted spans
                        # and original spans based on the begin index.
                        _insert_new_span(
                            entry_to_copy,
                            insert_ind,
                            inserted_annos,
                            new_pack,
                            spans,
                            new_spans,
                        )
                        insert_ind += 1

                    # Deletion
                    if orig_anno.tid in self._deleted_annos_id[pid]:
                        continue

                # Auto align the spans.
                span_new_begin: int = orig_anno.begin
                span_new_end: int = orig_anno.end

                if (
                    entry_to_copy == self.configs["augment_entry"]
                    or self._other_entry_policy[entry_to_copy] == "auto_align"
                ):
                    # Only inclusive when the entry is not augmented.
                    # E.g.: A Sentence include the inserted Token on the edge.
                    # E.g.: A Token shouldn't include a nearby inserted Token.
                    is_inclusive = (
                        entry_to_copy != self.configs["augment_entry"]
                    )
                    span_new_begin = modify_index(
                        orig_anno.begin, spans, new_spans, True, is_inclusive
                    )
                    span_new_end = modify_index(
                        orig_anno.end, spans, new_spans, False, is_inclusive
                    )

                new_anno = create_class_with_kwargs(
                    entry_to_copy,
                    {
                        "pack": new_pack,
                        "begin": span_new_begin,
                        "end": span_new_end,
                    },
                )
                new_pack.add_entry(new_anno)
                entry_map[orig_anno.tid] = new_anno.tid

            # Deal with spans after the last annotation in the original pack.
            if entry_to_copy == self.configs["augment_entry"]:
                while insert_ind < len(inserted_annos):
                    _insert_new_span(
                        entry_to_copy,
                        insert_ind,
                        inserted_annos,
                        new_pack,
                        spans,
                        new_spans,
                    )
                    insert_ind += 1

        # Iterate over and copy the links/groups in the datapack.
        for link in data_pack.get(Link):
            self._copy_link_or_group(link, entry_map, new_pack)
        for group in data_pack.get(Group):
            self._copy_link_or_group(group, entry_map, new_pack)

        self._data_pack_map[pid] = new_pack.pack_id
        self._entry_maps[pid] = entry_map
        return new_pack

    def _clear_states(self):
        r"""
        This function clears the states. It should be
        called after processing a multipack.
        """
        self._replaced_annos.clear()
        self._inserted_annos_pos_len.clear()
        self._deleted_annos_id.clear()
        self._data_pack_map.clear()
        self._entry_maps.clear()

    def _copy_link_or_group(
        self,
        entry: Union[Link, Group],
        entry_map: Dict[int, int],
        new_pack: DataPack,
    ) -> bool:
        r"""
        This function copies a Link/Group in the data pack.
        If the children Link/Group does not exist, it will recursively
        create the children Link/Group. If the children Annotation
        does not exist, it will abort and return False.
        Args:
            entry: The Link/Group in the original data pack to copy.
            entry_map: The dictionary mapping original entry to copied entry.
            new_pack: The new data pack, which is the destination of copy.
        Returns:
            A bool value indicating whether the copy happens.
        """

        # If the entry has been copied, return True.
        if entry.tid in entry_map:
            return True

        # The entry should be either Link or Group.
        is_link: bool = isinstance(entry, Link)

        # Get the children entries.
        children: List[Entry]
        if is_link:
            children = [entry.get_parent(), entry.get_child()]  # type: ignore
        else:
            children = entry.get_members()  # type: ignore

        # Copy the children entries.
        new_children: List[Entry] = []
        for child_entry in children:
            if isinstance(child_entry, (Link, Group)):
                # Recursively copy the children Links/Groups.
                if not self._copy_link_or_group(
                    child_entry, entry_map, new_pack
                ):
                    return False
            else:
                # Children Annotation must have been copied.
                if child_entry.tid not in entry_map:
                    return False
            new_child: Entry = new_pack.get_entry(entry_map[child_entry.tid])
            new_children.append(new_child)

        # Create the new entry and add to the new pack.
        new_entry: Entry
        if is_link:
            entry = cast(Link, entry)
            new_link_parent: Entry = new_children[0]
            new_link_child: Entry = new_children[1]
            new_entry = type(entry)(
                new_pack, new_link_parent, new_link_child  # type: ignore
            )
        else:
            entry = cast(Group, entry)
            new_entry = type(entry)(new_pack, new_children)  # type: ignore
        new_pack.add_entry(new_entry)
        entry_map[entry.tid] = new_entry.tid
        return True

    def _copy_multi_pack_link_or_group(
        self, entry: Union[MultiPackLink, MultiPackGroup], multi_pack: MultiPack
    ) -> bool:
        r"""
        This function copies a MultiPackLink/MultiPackGroup in the multipack.
        It could be used in tasks such as text generation, where
        MultiPackLink is used to align the source and target.
        Args:
            entry: The MultiPackLink/MultiPackGroup to copy.
            multi_pack: The multi_pack contains the input entry.
        Returns:
            A bool value indicating whether the copy happens.
        """
        # The entry should be either MultiPackLink or MultiPackGroup.
        is_link: bool = isinstance(entry, BaseLink)
        children: List[Entry]
        if is_link:
            children = [entry.get_parent(), entry.get_child()]  # type: ignore
        else:
            children = entry.get_members()  # type: ignore

        # Get the copied children entries.
        new_children: List[Entry] = []
        for child_entry in children:
            child_pack: DataPack = child_entry.pack
            child_pack_pid: int = child_pack.pack_id
            # The new pack should be present.
            if (
                child_pack_pid not in self._data_pack_map
                or child_pack_pid not in self._entry_maps
            ):
                return False
            new_child_pack: DataPack = multi_pack.get_pack_at(
                multi_pack.get_pack_index(self._data_pack_map[child_pack_pid])
            )
            # The new child entry should be present.
            if child_entry.tid not in self._entry_maps[child_pack_pid]:
                return False
            new_child_tid: int = self._entry_maps[child_pack_pid][
                child_entry.tid
            ]
            new_child_entry: Entry = new_child_pack.get_entry(new_child_tid)
            new_children.append(new_child_entry)

        # Create the new entry and add to the multi pack.
        new_entry: Entry
        if is_link:
            entry = cast(MultiPackLink, entry)
            new_link_parent: Entry = new_children[0]
            new_link_child: Entry = new_children[1]
            new_entry = type(entry)(
                multi_pack, new_link_parent, new_link_child  # type: ignore
            )
        else:
            entry = cast(MultiPackGroup, entry)
            new_entry = type(entry)(multi_pack, new_children)  # type: ignore
        multi_pack.add_entry(new_entry)
        return True

    def _augment(self, input_pack: MultiPack, aug_pack_names: List[str]):
        r"""
        This function calls the data augmentation ops and
        modifies the input in-place. This function only applies for
        replacement-based methods. The subclasses should override
        this function to implement other data augmentation methods, such
        as Easy Data Augmentation.
        Args:
            input_pack: The input MultiPack.
            aug_pack_names: The packs names for DataPacks to be augmented.
        """
        replacement_op = create_class_with_kwargs(
            self.configs["data_aug_op"],
            class_args={"configs": self.configs["data_aug_op_config"]},
        )
        augment_entry = get_class(self.configs["augment_entry"])

        for pack_name in aug_pack_names:
            data_pack: DataPack = input_pack.get_pack(pack_name)
            anno: Annotation
            for anno in data_pack.get(augment_entry):
                self._replace(replacement_op, anno)

    def _process(self, input_pack: MultiPack):
        # Get the pack names for augmentation.
        aug_pack_names: List[str] = []

        # Check if the DataPack exists.
        for pack_name in self.configs["augment_pack_names"].keys():
            if pack_name in input_pack.pack_names:
                aug_pack_names.append(pack_name)

        if len(self.configs["augment_pack_names"].keys()) == 0:
            # Augment all the DataPacks if not specified.
            aug_pack_names = list(input_pack.pack_names)

        self._augment(input_pack, aug_pack_names)
        new_packs: List[Tuple[str, DataPack]] = []
        for aug_pack_name in aug_pack_names:
            new_pack_name: str = self.configs["augment_pack_names"].get(
                aug_pack_name, "augmented_" + aug_pack_name
            )
            data_pack = input_pack.get_pack(aug_pack_name)
            new_pack = self._auto_align_annotations(
                data_pack=data_pack,
                replaced_annotations=self._replaced_annos[data_pack.pack_id],
            )
            new_packs.append((new_pack_name, new_pack))

        for new_pack_name, new_pack in new_packs:
            input_pack.add_pack_(new_pack, new_pack_name)

        # Copy the MultiPackLinks/MultiPackGroups
        for mpl in input_pack.get(MultiPackLink):
            self._copy_multi_pack_link_or_group(mpl, input_pack)
        for mpg in input_pack.get(MultiPackGroup):
            self._copy_multi_pack_link_or_group(mpg, input_pack)

        # Must be called after processing each multipack
        # to reset internal states.
        self._clear_states()

    @classmethod
    def default_configs(cls):
        """
        Returns:
            A dictionary with the default config for this processor.
        Following are the keys for this dictionary:
            - augment_entry:
                Defines the entry the processor will augment.
                It should be a full qualified name of the entry class.
                For example, "ft.onto.base_ontology.Sentence".
            - other_entry_policy:
                A dict specifying the policies for other entries.
                The key should be a full qualified class name.
                The policy(value of the dict) specifies how to process
                the corresponding entries after replacement.
                If the policy is "auto_align", the span of the entry
                will be automatically modified according to its original
                location. However, some spans might become invalid after
                the augmentation, for example, the tokens within a
                replaced sentence may disappear.
                Annotations not in the "other_entry_policy" will not
                be copied to the new data pack. The Links and Groups
                will be copied as well if the annotations they are
                attached to are copied.
                Example:
                    .. code-block:: python
                        'other_entry_policy': {
                            "kwargs": {
                                "ft.onto.base_ontology.Document": "auto_align",
                                "ft.onto.base_ontology.Sentence": "auto_align",
                            }
                        }
            - type:
                Should not modify this field, in order to use the kwargs.
            - data_aug_op:
                The data augmentation Op for the processor.
                It should be a full qualified class name.
                Example:
                    "forte.processors.data_augment.algorithms.
                    text_replacement_op.TextReplacementOp"
            - data_aug_op_config:
                The configuration for data augmentation Op.
                Example:
                    .. code-block:: python
                        'data_aug_op_config': {
                            'kwargs': {
                                'lang': 'en',
                                'use_gpu': False
                            }
                        }
            - augment_pack_names:
                A dict specifies the DataPacks to augment and their output
                names. It should be key-value pairs where the key is the
                input DataPack name, and the value is the output DataPack
                name after augmentation.
                If empty, all the DataPacks will be augmented, and the output
                names will be automatically generated by prepending
                an `'augmented_'` prefix.
                Example:
                    .. code-block:: python
                        'data_aug_op_config': {
                            'kwargs': {
                                'src': 'aug_src',
                                'tgt': 'aug_tgt'
                            }
                        }
        """
        return {
            "augment_entry": "ft.onto.base_ontology.Sentence",
            "other_entry_policy": {},
            "type": "data_augmentation_op",
            "data_aug_op": "",
            "data_aug_op_config": {},
            "augment_pack_names": {},
            "@no_typecheck": [
                "other_entry_policy",
                "data_aug_op_config",
                "augment_pack_names",
            ],
        }



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=attribute-defined-outside-init
# pylint: disable=useless-super-delegation
"""
An elastic search indexer needs to be created first
in order to perform data selection.
Refer to examples/data_augmentation/data_select_index_pipeline.py
for indexer creation.
"""
from abc import ABC
from typing import Iterator, Any, Dict, Optional

from forte.common.configuration import Config
from forte.common.resources import Resources
from forte.data.base_reader import PackReader
from forte.data.data_pack import DataPack

__all__ = [
    "BaseElasticSearchDataSelector",
    "RandomDataSelector",
    "QueryDataSelector",
]

from forte.utils import create_class_with_kwargs


class BaseDataSelector(PackReader, ABC):
    r"""A base data selector for data augmentation.
    It is a reader that selects a subset from the dataset and yields datapacks.
    """


class BaseElasticSearchDataSelector(BaseDataSelector):
    r"""The base elastic search indexer for data selector. This class creates
    an :class:`~fortex.elastic.elastic_indexer.ElasticSearchIndexer`
    and searches for documents according to the user-provided search keys.
    Currently supported search criteria: random-based and query-based. It
    then yields the corresponding datapacks of the selected documents.
    """

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)
        self.index = create_class_with_kwargs(
            self.configs.indexer_class,
            class_args={"config": self.configs.index_config},
        )

    def _create_search_key(self, data: Optional[str]) -> Dict[str, Any]:
        raise NotImplementedError

    def _collect(self, *args, **kwargs) -> Iterator[str]:
        raise NotImplementedError

    def _parse_pack(self, pack_info: str) -> Iterator[DataPack]:
        pack: DataPack = DataPack.from_string(pack_info)  # type: ignore
        yield pack

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        config = super().default_configs()
        config.update(
            {
                "indexer_class": "fortex.elastic.elastic_indexer.ElasticSearchIndexer",
                "index_config": {
                    "index_name": "elastic_indexer",
                    "hosts": "localhost:9200",
                    "algorithm": "bm25",
                },
            }
        )
        return config


class QueryDataSelector(BaseElasticSearchDataSelector):
    def _collect(self, *args, **kwargs) -> Iterator[str]:
        # pylint: disable = unused-argument
        r"""Iterator over query text files in the data_source.
        Then search for documents using indexer.

        Args:
            args: args[0] is the path to the query file
            kwargs:

        Returns: Selected document's original datapack.
        """
        data_path: str = args[0]
        with open(data_path, "r", encoding="utf-8") as file:
            for line in file:
                query: Dict = self._create_search_key(line.strip())
                results = self.index.search(query)
                hits = results["hits"]["hits"]

                for _, hit in enumerate(hits):
                    document = hit["_source"]
                    yield document["pack_info"]

    def _create_search_key(self, data: str) -> Dict[str, Any]:  # type: ignore
        r"""Create a search dict for elastic search indexer.
        Args:
             text: str
                A string which will be looked up for in the corpus under field
                name `field`. It can be passed in `config` during initialize.
                If `config` does not contain `field`, we set it to "content".

        Returns: A dict that specifies the query match field.
        """
        return {
            "query": {"match": {self.configs["field"]: data}},
            "size": self.configs["size"],
        }

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        config = super().default_configs()
        config.update(
            {
                "size": 1000,
                "field": "content",
            }
        )
        return config


class RandomDataSelector(BaseElasticSearchDataSelector):
    def _collect(self, *args, **kwargs) -> Iterator[str]:
        # pylint: disable = unused-argument
        r"""random select `size` documents from the indexer.
        Returns: Selected document's original datapack.
        """
        query: Dict = self._create_search_key()
        results = self.index.search(query)
        hits = results["hits"]["hits"]

        for _, hit in enumerate(hits):
            document = hit["_source"]
            yield document["pack_info"]

    def _create_search_key(self) -> Dict[str, Any]:  # type: ignore
        return {
            "size": self.configs["size"],
            "query": {
                "function_score": {
                    "functions": [
                        {
                            "random_score": {
                                "seed": "1477072619038",
                                "field": "_seq_no",
                            }
                        }
                    ]
                }
            },
        }

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        config = super().default_configs()
        config.update({"size": 1000000})
        return config



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Dict, Union, List, Optional, Tuple, Type

from forte.common import ProcessorConfigError
from forte.common.configuration import Config
from forte.data.base_extractor import BaseExtractor
from forte.data.converter.feature import Feature
from forte.data.data_pack import DataPack
from forte.data.ontology import Annotation, Link
from forte.data.ontology.core import Entry

__all__ = ["LinkExtractor"]

from forte.utils import get_class


def get_index(
    pack: DataPack, index_entries: List[Annotation], context_entry: Annotation
):
    founds = []
    for i, entry in enumerate(index_entries):
        if pack.covers(context_entry, entry):
            founds.append(i)
    return [founds[0], founds[-1] + 1]


class LinkExtractor(BaseExtractor):
    """
    This extractor extracts relation type features from data packs. This
    extractor expects the parent and child of the relation to be Annotation
    entries.
    """

    def initialize(self, config: Union[Dict, Config]):
        # pylint: disable=attribute-defined-outside-init
        super().initialize(config)

        if self.config.attribute is None:
            raise ProcessorConfigError(
                "'attribute' is required in this extractor."
            )
        if self.config.index_annotation is None:
            raise ProcessorConfigError(
                "'index_annotation' is required in this extractor."
            )
        if self.config.entry_type is None:
            raise ProcessorConfigError(
                "'entry_type' is required in this extractor."
            )
        else:
            self._entry_class: Type[Link] = get_class(self.config.entry_type)

            if not issubclass(self._entry_class, Link):
                raise ProcessorConfigError(
                    "`entry_class` to this extractor " "must be a Link tpe."
                )

            self._parent_class: Type[Annotation] = self._entry_class.ParentType
            if not issubclass(self._parent_class, Annotation):
                raise ProcessorConfigError(
                    f"The parent class of the provided {self.config.entry_type}"
                    " must be an Annotation."
                )

            self._child_class: Type[Annotation] = self._entry_class.ChildType
            if not issubclass(self._child_class, Annotation):
                raise ProcessorConfigError(
                    f"The child class of the provided {self.config.entry_type}"
                    " must be an Annotation."
                )

    @classmethod
    def default_configs(cls):
        r"""Returns a dictionary of default hyper-parameters.

        Here:

        - "`entry_type`": The target relation entry type, should be a Link entry.
        - "`attribute`": The attribute of the relation to extract.
        - "`index_annotation`": The annotation object used to index the
          head and child node of the relations.
        """
        config = super().default_configs()
        config.update(
            {
                "entry_type": None,
                "attribute": None,
                "index_annotation": None,
            }
        )
        return config

    def update_vocab(
        self, pack: DataPack, context: Optional[Annotation] = None
    ):
        """
        Update values of relation attributes to the vocabulary.

        Args:
            pack: The input data pack.
            context: The context is an Annotation entry where
                features will be extracted within its range. If None, then the
                whole data pack will be used as the context. Default is None.
        Returns:
            None
        """

        entry: Entry
        for entry in pack.get(self.config.entry_type, context):
            attribute = getattr(entry, self.config.attribute)
            self.add(attribute)

    def extract(
        self, pack: DataPack, context: Optional[Annotation] = None
    ) -> Feature:
        """Extract link data as features from the context.

        Args:
            pack: The input data pack that contains the features.
            context: The context is an Annotation entry where
                features will be extracted within its range. If None, then the
                whole data pack will be used as the context. Default is None.

        Returns:

        """
        index_annotations: List[Annotation] = list(
            pack.get(self.config.index_annotation, context)
        )

        parent_nodes: List[Annotation] = []
        child_nodes: List[Annotation] = []
        relation_atts = []

        r: Link
        for r in pack.get(self.config.entry_type, context):
            parent_nodes.append(r.get_parent())  # type: ignore
            child_nodes.append(r.get_child())  # type: ignore

            raw_att = getattr(r, self.config.attribute)
            relation_atts.append(
                self.element2repr(raw_att) if self.vocab else raw_att
            )

        parent_unit_span = []
        child_unit_span = []

        for p, c in zip(parent_nodes, child_nodes):
            parent_unit_span.append(get_index(pack, index_annotations, p))
            child_unit_span.append(get_index(pack, index_annotations, c))

        meta_data = {
            "parent_unit_span": parent_unit_span,
            "child_unit_span": child_unit_span,
            "pad_value": self.get_pad_value(),
            "dim": 1,
            "dtype": int if self.vocab else str,
        }

        return Feature(data=relation_atts, metadata=meta_data, vocab=self.vocab)

    def add_to_pack(
        self,
        pack: DataPack,
        predictions: List[Tuple[Tuple[int, int], Tuple[int, int], int]],
        context: Optional[Annotation] = None,
    ):
        """
        Convert prediction back to Links inside the data pack.

        Args:
            pack: The datapack to add predictions back.
            predictions: This is the output of the model, it is a
                triplet, the first element shows the parent, the second
                element shows the child. These two are indexed by the
                `index_annotation` of this extractor. The last element is the
                index of the relation attribute.
            context: The context is an Annotation
                entry where predictions will be added to. This has the same
                meaning with `context` as in
                :meth:`~forte.data.base_extractor.BaseExtractor.extract`.
                If None, then the whole data pack will be used as the
                context. Default is None.
        """
        index_entries: List[Annotation] = list(
            pack.get(self.config.index_annotation, context)
        )

        for parent, child, rel_index in predictions:
            parent_begin_entry_index, parent_end_entry_index = parent
            child_begin_entry_index, child_end_entry_index = child

            parent_start = index_entries[parent_begin_entry_index].begin
            parent_end = index_entries[parent_end_entry_index].end

            child_start = index_entries[child_begin_entry_index].begin
            child_end = index_entries[child_end_entry_index].end
            rel_value = self.id2element(rel_index)

            child_anno = self._child_class(
                pack, child_start, child_end  # type:ignore
            )

            parent_anno = self._parent_class(
                pack, parent_start, parent_end  # type:ignore
            )

            link = self._entry_class(
                pack, parent_anno, child_anno  # type:ignore
            )
            setattr(link, self.config.attribute, rel_value)



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This class contains readers to read from the Stave annotation tool.

The Stave annotation tool can be found here: https://github.com/asyml/stave
"""
import logging
import sqlite3
from typing import Iterator, Dict, Optional

from forte.common import Resources, ProcessorConfigError
from forte.common.configuration import Config
from forte.data.base_pack import BasePack
from forte.data.data_pack import DataPack
from forte.data.multi_pack import MultiPack
from forte.data.base_reader import PackReader
from forte.data.readers.deserialize_reader import MultiPackDeserializerBase

__all__ = [
    "StaveMultiDocSqlReader",
    "StaveDataPackSqlReader",
]


def load_all_datapacks(
    conn, pack_table_name: str, pack_col: int
) -> Dict[int, DataPack]:
    """
    Load all the data packs from the table given a sqlite connection to the
      Stave database.

    Args:
        conn:  The sqlite database connection.
        pack_table_name: The name of the pack table.
        pack_col: The column number to retrieve the actual pack content.

    Returns:
        A dictionary contains all the datapacks.

    """
    c = conn.cursor()
    data_packs: Dict[int, DataPack] = {}
    for val in c.execute(f"SELECT * FROM {pack_table_name}"):
        pack: DataPack = DataPack.from_string(val[pack_col])  # type:ignore
        # Currently assume we do not have access to the id in the database,
        #  once we update all Stave db format, we can add the real id.
        data_packs[pack.pack_id] = pack
    return data_packs


class StaveMultiDocSqlReader(MultiPackDeserializerBase):
    """
    This reader reads multi packs from Stave's database schema.

    Users of this reader need to first make sure to build the ontology classes
    based on the ontology from the database.

    .. code-block:: python

        pl = Pipeline()
        pl.set_reader(StaveMultiDocSqlReader, config={
            "stave_db_path": self.sql_db,
        })

    The current Stave master does not provide a project key for multi doc, this
    implementation does not use the project key.

    Here, note that each project may have a different ontology system, thus
    users of this project need to make sure the classes correspond to the
    ontology are in place.

    .. note::

        Stave is a annotation interface built on Forte's format:
        - https://github.com/asyml/stave
    """

    def initialize(self, resources: Resources, configs: Config):
        # pylint: disable=attribute-defined-outside-init
        super().initialize(resources, configs)

        if not configs.stave_db_path:
            raise ProcessorConfigError(
                "The database path to stave is not specified."
            )

        self.conn = sqlite3.connect(configs.stave_db_path)
        self.data_packs: Dict[int, DataPack] = load_all_datapacks(
            self.conn, configs.datapack_table, configs.pack_content_col
        )

    def _get_multipack_content(self) -> Iterator[str]:  # type: ignore
        c = self.conn.cursor()
        for value in c.execute(
            f"SELECT textPack FROM {self.configs.multipack_table}"
        ):
            yield value[0]

    def _parse_multi_pack(self, multi_pack_source: str) -> MultiPack:
        return BasePack.from_string(multi_pack_source)  # type: ignore

    def _get_pack(self, pack_id: int) -> Optional[DataPack]:
        return self.data_packs.get(pack_id, None)

    @classmethod
    def default_configs(cls):
        """
        The default configurations for this reader, the default value of the
        configurations are as followed:

        .. code-block:: python

            {
                "stave_db_path": None,
                "multipack_table": 'stave_backend_crossdoc',
                "multipack_content_col": 2,
                "datapack_table": 'stave_backend_document',
                "pack_content_col": 2,
            }

        Here:

        `"stave_db_path"`: str
            Path to the database file.
        `"multipack_table"`: str
            The table name that contains the multi pack content.
        `"multipack_content_col"`: int
            Indicate the column of the `multipack_table` that contains the
            multi pack content.
        `"datapack_table"`: str
            The table name that contains the data pack content.
        `"pack_content_col"`: str
            The col that contains the data pack content.

        Returns: The default configs of this reader.

        """
        config = super().default_configs()
        config.update(
            {
                "stave_db_path": None,
                "multipack_table": "stave_backend_crossdoc",
                "multipack_content_col": 2,
                "datapack_table": "stave_backend_document",
                "pack_content_col": 2,
            }
        )
        return config


class StaveDataPackSqlReader(PackReader):
    """
    This reader reads data packs from Stave's database.

    Users of this reader need to first make sure to build the ontology classes
    based on the ontology from the database.

    .. code-block:: python

        pl = Pipeline()
        pl.set_reader(StaveMultiDocSqlReader, config={
            "stave_db_path": self.sql_db,
            "target_project_name": project_name
        })

    Here, note that each project may have a different ontology system, thus
    users of this project need to make sure the classes correspond to the
    ontology are in place.

    If you are able to use a shared ontology for all projects, you may also
    read all data packs without specifying the project name.

    .. code-block:: python

        pl = Pipeline()
        pl.set_reader(StaveMultiDocSqlReader, config={
            "stave_db_path": self.sql_db
        })


    .. note::

        Stave is a annotation interface built on Forte's format:
        - https://github.com/asyml/stave
    """

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)

        if not configs.stave_db_path:
            raise ProcessorConfigError(
                "The database path to stave is not specified."
            )

        if not configs.datapack_table:
            raise ProcessorConfigError(
                "The table name that stores the data pack is not stored."
            )

        if not configs.target_project_name:
            logging.info(
                "No project specified, will attempt to read all proejcts."
            )

    def _collect(self) -> Iterator[str]:  # type: ignore
        # pylint: disable=attribute-defined-outside-init
        self.conn = sqlite3.connect(self.configs.stave_db_path)
        c = self.conn.cursor()

        pack: str = self.configs.datapack_table
        project: str = self.configs.project_table

        if self.configs.target_project_name is None:
            # Read all documents in the database.
            query = f"SELECT textPack FROM {pack}"
        else:
            # Read the specific project.
            query = (
                f"SELECT textPack FROM {pack}, {project} "
                f"WHERE {pack}.project_id = {project}.id "
                f'AND {project}.name = "{self.configs.target_project_name}"'
            )

        for value in c.execute(query):
            yield value[0]

    def _parse_pack(self, pack_str: str) -> Iterator[DataPack]:
        yield DataPack.from_string(pack_str)  # type: ignore

    @classmethod
    def default_configs(cls):
        """
        The default configurations in this project, the default value of the
        configurations are as followed:

        .. code-block:: python

            {
                "stave_db_path": None,
                "datapack_table": 'stave_backend_document',
                "pack_content_col": 2,
                "project_table": "stave_backend_project",
                "target_project_name": None,
            }

        Here:

        `"stave_db_path"`: str
            Path to the database file.
        `"datapack_table"`: str
            The table name that contains the data pack documents.
        `"pack_content_col"`: int
            Indicate the column of the table that contains the data
              pack content.
        `"project_table"`: str
            The table that contains the project information.
        `"target_project_name"`: str
            The table name that contains the project information.

        Returns: The default configs of this reader.

        """
        config = super().default_configs()
        config.update(
            {
                "stave_db_path": None,
                "datapack_table": "stave_backend_document",
                "pack_content_col": 2,
                "project_table": "stave_backend_project",
                "target_project_name": None,
            }
        )
        return config



# Copyright 2021 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A StaveProcessor is introduced to enable immediate visualization of forte
pipeline result. It supports serving Stave instance with StaveViewer.
Forte users can plug it into the pipeline to easily visualize datapacks
with annotations. It is also highly configurable for users to change port
number, host name, layout, etc.

Package Requirements:
    forte
    stave
"""

import os
import logging
import collections
from typing import Dict, Set, Any
from forte.utils import create_import_error_msg
from forte.common import Resources, ProcessorConfigError
from forte.common.configuration import Config
from forte.data.data_pack import DataPack
from forte.data.ontology.code_generation_objects import search
from forte.processors.base import PackProcessor

try:
    from stave_backend.lib.stave_viewer import StaveViewer
    from stave_backend.lib.stave_project import StaveProjectWriter
except ImportError as e:
    raise ImportError(
        create_import_error_msg("stave", "stave", "stave processor")
    ) from e


logger = logging.getLogger(__name__)

__all__ = ["StaveProcessor"]


class StaveProcessor(PackProcessor):
    r"""
    ``StaveProcessor`` provides easy visualization for forte users. We can
    visualize datapack with annotations by inserting it into the forte
    pipeline without affecting the original functionalities.

    ``StaveProcessor`` requires an ontology file being passed to the pipeline
    by setting the input parameter ``ontology_file``. Otherwise a
    ``ProcessorConfigError`` will be raised. It then generates default
    configurations based on the input ontology to start a stave instance
    without any additional specification by users.
    Example usage:

    .. code-block:: python

        Pipeline(ontology_file="ontology/path") \
            .set_reader(plaintext_reader(), {"input_path":"some/path"}) \
            .add(StaveProcessor())

    After initialized, ``StaveProcessor`` will create a project directory
    (or use an existing directory specified in ``project_path``). Metadata
    and textpacks will be dumped into the direcotry.

    ``StaveProcessor`` is also highly customizable for users to set up. Users
    may configure port number, server host, project name, etc.
    Example usage:

    .. code-block:: python

        Pipeline(ontology_file="ontology/path") \
            .set_reader(plaintext_reader(), {"input_path":"some/path"}) \
            .add(StaveProcessor(), configs={
                "port": 8880,
                "project_name": "serialization_pipeline_test"
            })

    Users can modify project configs by changing the ``project_configs`` field.
    Example usage:

    .. code-block:: python

        Pipeline(ontology_file="ontology/path") \
            .set_reader(plaintext_reader(), {"input_path":"some/path"}) \
            .add(StaveProcessor(), configs={
                "port": 8879,
                "project_configs": {
                    # Configure Stave layout. Replace the normal annotation
                    # viewer "default-nlp" with a dialogue box.
                    "layoutConfigs": {
                        "center-middle": "DialogueBox"
                    }
                }
            })
    """

    def __init__(self):
        super().__init__()
        self._project_id: int = -1
        self._viewer: StaveViewer
        self._project_writer: StaveProjectWriter

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)

        # Validate multi_pack project config:
        #   A `multi_pack` project must have `multi_ontology` set.
        if self.configs.project_type != "single_pack" and (
            self.configs.project_type != "multi_pack"
            or self.configs.multi_ontology is None
        ):
            raise ProcessorConfigError("Invalid project type configuration.")

        # Generate default configurations
        self.configs.project_configs = Config(
            hparams=self.configs.project_configs,
            default_hparams=self._default_project_configs(),
        )
        self.configs.multi_ontology = self.configs.multi_ontology or Config(
            {}, {}
        )
        self.configs.project_path = os.path.abspath(
            self.configs.project_path or self.configs.project_name
        )

        self._viewer = StaveViewer(
            project_path=self.configs.project_path,
            host=self.configs.host,
            port=self.configs.port,
            thread_daemon=self.configs.server_thread_daemon,
        )

        #  Write meta data to project folder
        self._project_writer = StaveProjectWriter(
            project_path=self.configs.project_path,
            project_name=self.configs.project_name,
            project_type=self.configs.project_type,
            ontology=self.resources.get("onto_specs_dict"),
            project_configs=self.configs.project_configs.todict(),
            multi_ontology=self.configs.multi_ontology.todict(),
        )

    def _process(self, input_pack: DataPack):

        if not self._viewer.server_started:
            self._viewer.run()

        if self._viewer.server_started:
            textpack_id = self._project_writer.write_textpack(
                input_pack.pack_name
                if self.configs.use_pack_name
                else input_pack.pack_id,
                input_pack.to_string(),
            )
            if textpack_id == 0:
                self._viewer.open()

    def _default_project_configs(self):
        # pylint: disable=line-too-long
        """
        Create default project configuration based on ontology.
        This is translated from JavaScript function `createDefaultConfig` in
        https://github.com/asyml/stave/blob
        /d82383de3d74bf09c0d30f33d8a902595f5aff80/src/app/pages/Projects.tsx
        #L140

        Returns:
            configs: A dictionary with the default config for project.

        """
        # pylint: enable=line-too-long

        if not (
            self.resources.contains("onto_specs_dict")
            and self.resources.contains("merged_entry_tree")
        ):
            raise ProcessorConfigError(
                "onto_specs_dict/merged_entry_tree is not set in resources."
            )
        ontology = self.resources.get("onto_specs_dict")
        entry_tree = self.resources.get("merged_entry_tree")

        configs: Dict[str, Any] = {
            "legendConfigs": {},
            "scopeConfigs": {},
            "layoutConfigs": {
                "center-middle": "default-nlp",
                "left": "default-meta",
                "right": "default-attribute",
                "center-bottom": "disable",
            },
            "remoteConfigs": {
                "pipelineUrl": "",
                "doValidation": False,
                "expectedName": "",
                "inputFormat": "string",
                "expectedRecords": {},
            },
        }

        # Create legend configs
        legend_configs: Dict[str, Any] = {}
        entry_name_set: Set[str] = set()
        for entry in ontology["definitions"]:
            entry_name = entry["entry_name"]
            entry_name_set.add(entry_name)
            legend_configs[entry_name] = {
                "is_selected": False,
                "is_shown": True,
            }
            if "attributes" in entry and len(entry["attributes"]) > 0:
                attributes_configs = {}
                for attribute in entry["attributes"]:
                    if attribute["type"] == "str":
                        attributes_configs[attribute["name"]] = False
                legend_configs[entry_name]["attributes"] = attributes_configs
        configs["legendConfigs"] = legend_configs

        # Find all subclass of `forte.data.ontology.top.Annotation` and
        # update `scopeConfigs` accordingly.
        queue = collections.deque(
            [search(entry_tree.root, "forte.data.ontology.top.Annotation")]
        )
        while queue:
            size = len(queue)
            for _ in range(size):
                node = queue.pop()
                if node.name in entry_name_set:
                    configs["scopeConfigs"][node.name] = False
                for entry in node.children:
                    queue.appendleft(entry)
        return configs

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        """
        This defines a basic config structure for StaveProcessor.
        Following are the keys for this dictionary:

            - ``project_path``: Path to the project directory for rendering.
              Default to None, which creates a folder using ``project_name``.
            - ``port``: Port number for Stave server. Default value is `8888`.
            - ``host``: Host name for Stave server. Default value is
              `"localhost"`.
            - ``project_type``: `"single_pack\"` (default) or `\"multi_pack"`.
            - ``project_name``: Project name displayed on Stave. Default name
              is `"Auto generated project"`.
            - ``multi_ontology``: A dictionary for multi_pack ontology
              Default to `None`.
            - ``project_configs``: Project configurations. Default to `None`.
            - ``server_thread_daemon``: Sets whether the server thread is
              daemonic. Default to `False`.
            - ``use_pack_name``: Use ``pack_name`` to name the textpack being
              saved to project path in viewer mode. If `False`, will use
              ``pack_id`` for naming. Default to False.

        Returns:
            dict: A dictionary with the default config for this processor.
        """
        return {
            "project_path": None,
            "port": 8888,
            "host": "localhost",
            "project_type": "single_pack",
            "project_name": "Auto generated project",
            "multi_ontology": None,
            "project_configs": None,
            "server_thread_daemon": False,
            "use_pack_name": False,
        }



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=attribute-defined-outside-init
import os
from typing import Dict, Any
from forte.utils import create_import_error_msg
from forte.common.configuration import Config
from forte.common.resources import Resources
from forte.data.multi_pack import MultiPack
from forte.data.ontology import Query
from forte.processors.base import MultiPackProcessor
from forte.processors.ir.bert.bert_ranker import (
    BERTClassifier,
    BERTEncoder,
)

try:
    import torch
except ImportError as e1:
    raise ImportError(
        create_import_error_msg("torch", "ir", "Information Retrieval supports")
    ) from e1


__all__ = ["BertRerankingProcessor"]


class BertRerankingProcessor(MultiPackProcessor):
    def initialize(self, resources: Resources, configs: Config):
        self.resources = resources
        self.config = Config(configs, self.default_configs())

        # TODO: At the time of writing, no way in texar to set encoder in
        # `texar.torch.modules.classifiers.BertClassifier`. Should not ideally
        # be changing a private variable.
        # pylint: disable=protected-access
        BERTClassifier._ENCODER_CLASS = BERTEncoder
        # pylint: enable=protected-access

        cache_dir = os.path.join(
            os.path.dirname(__file__), self.config.model_dir
        )

        self.device = (
            torch.device("cuda:0")
            if torch.cuda.is_available()
            else torch.device("cpu")
        )

        self.model = BERTClassifier(
            pretrained_model_name=self.config.pretrained_model_name,
            cache_dir=cache_dir,
            hparams=self.config,
        ).to(self.device)
        try:
            from texar.torch.data import (  # pylint: disable=import-outside-toplevel
                BERTTokenizer,
            )
        except ImportError as e:
            raise ImportError(
                create_import_error_msg("texar-pytorch", "ir", "IR support")
            ) from e
        self.tokenizer = BERTTokenizer(
            pretrained_model_name=self.config.pretrained_model_name,
            cache_dir=cache_dir,
            hparams=None,
        )

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        return {
            "size": 5,
            "query_pack_name": "query",
            "field": "content",
            "pretrained_model_name": "bert-large-uncased",
            "model_dir": os.path.join(os.path.dirname(__file__), "models"),
            "max_seq_length": 512,
        }

    def _process(self, input_pack: MultiPack):
        max_len = self.config.max_seq_length
        query_pack_name = self.config.query_pack_name

        query_pack = input_pack.get_pack(self.config.query_pack_name)
        query_entry = list(query_pack.get(Query))[0]
        query_text = query_pack.text

        packs = {}
        for doc_id in input_pack.pack_names:
            if doc_id == query_pack_name:
                continue

            pack = input_pack.get_pack(doc_id)
            document_text = pack.text

            # BERT Inference
            input_ids, segment_ids, input_mask = [
                torch.LongTensor(item).unsqueeze(0).to(self.device)
                for item in self.tokenizer.encode_text(
                    query_text, document_text, max_len
                )
            ]

            seq_length = (input_mask == 1).sum(dim=-1)
            logits, _ = self.model(input_ids, seq_length, segment_ids)
            preds = torch.nn.functional.softmax(torch.Tensor(logits), dim=1)

            score = preds.detach().tolist()[0][1]

            query_entry.update_results({doc_id: score})
            packs[doc_id] = pack



#  Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy
from typing import List, Any, Tuple, Union, Dict, Optional


from forte.common import ValidationError
from forte.data.vocabulary import Vocabulary

__all__ = ["Feature"]


class Feature:
    """
    This class represents a type of feature for a single data instance. The
    Feature can be multiple dimensions. It has methods to do padding and
    retrieve the actual multi-dimension data.

    Args:
        data: A list of features, where each feature can be the value or
            another list of features. Typically this should be the output from
            :meth:`extract` in
            :class:`~forte.data.base_extractor.BaseExtractor`.

        metadata: A dictionary storing meta-data for this feature.
            Mandatory fields includes: `dim`, `dtype`.

            - `dim` indicates the total number of dimension for this
              feature.
            - `dtype` is the value type. For example, it can be `torch.long`.

        vocab: An optional fields about the
            :class:`~forte.data.vocabulary.Vocabulary` used to build this
            feature.

    Please refer to :meth:`data` for the typical usage of this class.
    """

    def __init__(
        self, data: List, metadata: Dict, vocab: Optional[Vocabulary] = None
    ):
        self._meta_data: Dict = metadata
        self._validate_metadata()

        # Pad that will be used to do padding.
        self._pad_value: Union[int, List] = self._meta_data["pad_value"]
        self._dim: int = self._meta_data["dim"]
        self._dtype = self._meta_data["dtype"]

        # Indicating whether current Feature is the inner most feature.
        self._leaf_feature: bool = self._dim == 1
        # Only leaf feature has actual `data`
        self._data: Union[None, List] = None
        # Only non-leaf features have `sub_features`. It can be considered as
        # a list of Feature instances
        self._sub_features: List = []
        # The elements of mask will indicate whether the corresponding value
        # in the data is the actual value or padded data. `mask` will only
        # indicate feature along current dimension. Sub-dimension mask will
        # be stored inside sub features in `sub_features` It will be updated
        # when the method `pad` is called.
        self._mask: List = []

        self._vocab: Optional[Vocabulary] = vocab

        self._parse_sub_features(data)
        self._validate_input()

    def _validate_metadata(self):
        necessary_fields = ["dim", "dtype"]
        for field in necessary_fields:
            if field not in self._meta_data:
                raise ValidationError(
                    "Field not found in metadata: {}".format(field)
                )

        if self.need_pad and "pad_value" not in self._meta_data:
            raise ValidationError("need_pad is True but no pad_value is given")

    def _validate_input(self):
        """
        Validate input parameters based on some pre-conditions.
        """
        if self.leaf_feature and self._data is None:
            raise ValidationError("Leaf feature should contain actual data.")
        if not self.leaf_feature and self._data:
            raise ValidationError(
                "Non-leaf feature should not contain actual data."
            )
        if not self.leaf_feature and not self._sub_features:
            raise ValidationError(
                "Non-leaf feature should contain sub features."
            )
        if self._dim < 1:
            raise ValidationError("The `dim` in meta should be at least 1.")

    def _parse_sub_features(self, data):
        """
        If current feature is the leaf feature, store the input data. Otherwise,
        parse the data into sub features represented as a list of features.
        Meanwhile, update the mask list.
        Args:
            data (List):
                A list of features, where each feature can be the value or
                another list of features.
        """
        if self.leaf_feature:
            self._data: List = data
        else:
            self._sub_features: List = []
            for sub_data in data:
                sub_metadata = deepcopy(self._meta_data)
                sub_metadata["dim"] = sub_metadata["dim"] - 1
                self._sub_features.append(
                    Feature(
                        data=sub_data, metadata=sub_metadata, vocab=self._vocab
                    )
                )

        self._mask: List = [1] * len(data)

    @property
    def leaf_feature(self) -> bool:
        """
        Returns:
            True if current feature is leaf feature. Otherwise, False.
        """
        return self._leaf_feature

    @property
    def dtype(self):
        """
        Returns:
            The data type of this feature.
        """
        return self._dtype

    @property
    def sub_features(self) -> List["Feature"]:
        """
        Returns:
            A list of sub features. Raise exception if current feature is the
            leaf feature.
        """
        if self._leaf_feature:
            raise ValidationError("Leaf feature does not have sub features")
        if self._dim <= 1:
            raise ValidationError(
                "Non-leaf feature should have as least 2 dimension"
            )

        return self._sub_features

    @property
    def meta_data(self) -> Dict:
        """
        Returns:
            A `Dict` of meta data describing this feature.
        """
        return self._meta_data

    @property
    def vocab(self) -> Optional[Vocabulary]:
        """
        Returns:
            The :class:`~forte.data.vocabulary.Vocabulary` used to build this
            feature.
        """
        return self._vocab

    @property
    def dim(self) -> int:
        """
        Returns:
            The dimension of this feature.
        """
        return self._dim

    @property
    def need_pad(self) -> bool:
        """
        Returns:
            Whether the Feature need to be padded.
        """
        return (
            self._meta_data["need_pad"]
            if "need_pad" in self._meta_data
            else True
        )

    def __len__(self):
        return len(self._data) if self.leaf_feature else len(self._sub_features)

    def pad(self, max_len: int):
        """
        Pad the current feature dimension with the given `max_len`. It will use
        `pad_value` to do the padding.

        Args:
            max_len (int):
                The padded length.
        """
        if len(self) > max_len:
            raise ValidationError(
                "Feature length should not exceed given max_len"
            )

        for _ in range(max_len - len(self)):
            if self.leaf_feature:
                if self._data is None:
                    raise ValueError(
                        "Invalid internal state: leaf_feature "
                        "does not have actual data"
                    )
                self._data.append(self._pad_value)
            else:
                sub_metadata = deepcopy(self._meta_data)
                sub_metadata["dim"] = sub_metadata["dim"] - 1
                self._sub_features.append(
                    Feature(data=[], metadata=sub_metadata, vocab=self._vocab)
                )
            self._mask.append(0)

    @property
    def data(self) -> Tuple[List[Any], List[Any]]:
        """
        It will return the actual data stored. Internally, it will recursively
        retrieve data from inner dimension features. Meanwhile, it will also
        return a list of masks representing the mask along different dimensions.

        Returns:
            A `Tuple` where

            The first element is the actual data representing this feature.

            The second element is a list of masks. masks[i] in this list
            represents the mask along i-th dimension.

        Here are some examples for how the padding works:

        Example 1 (1-dim feature, no padding):

        .. code-block:: python

            data = [2,7,8]
            meta_data = {
                "pad_value": 0
                "dim": 1
                "dtype": torch.long
            }
            feature = Feature(data, meta_data=meta_data)

            data, masks = feature.data

            # data is:
            # [2,7,8]

            # masks is:
            # [
            #   [1,1,1]
            # ]

        Example 2 (1-dim feature, scalar padding):

        .. code-block:: python

            data = [2,7,8]
            meta_data = {
                "pad_value": 0
                "dim": 1
                "dtype": torch.long
            }
            feature = Feature(data, meta_data=meta_data)

            feature.pad(max_len=4)

            data, masks = feature.data

            # data is:
            # [2,7,8,0]

            # masks is:
            # [
            #   [1,1,1,0]
            # ]

        Example 3 (2-dim feature, scalar padding):

        .. code-block:: python

            data = [[1,2,5], [3], [1,5]]
            meta_data = {
                "pad_value": 0
                "dim": 2
                "dtype": torch.long
            }
            feature = Feature(data, meta_data=meta_data)

            feature.pad(max_len=4)
            for sub_feature in feature.sub_features:
                sub_feature.pad(max_len=3)

            data, masks = feature.data

            # data is:
            # [[1,2,5], [3,0,0], [1,5,0], [0,0,0]]

            # masks is:
            # [
            #   [1,1,1,0],
            #   [[1,1,1], [1,0,0], [1,1,0], [0,0,0]]
            # ]

        Example 4 (1-dim feature, vector padding):

        .. code-block:: python

            data = [[0,1,0],[1,0,0]]
            meta_data = {
                "pad_value": [0,0,1]
                "dim": 1
                "dtype": torch.long
            }
            feature = Feature(data, meta_data=meta_data)

            feature.pad(max_len=3)

            data, masks = feature.data

            # data is:
            # [[0,1,0], [1,0,0], [0,0,1]]

            # masks is:
            # [
            #  [1,1,0]
            # ]
        """
        if self.leaf_feature:
            if self._data is None:
                raise ValueError(
                    "Invalid internal state: leaf_feature "
                    "does not have actual data"
                )
            return self._data, [self._mask]
        else:
            unroll_features: List = []
            sub_stack_masks: List = []

            for feature in self.sub_features:
                sub_unroll_features, sub_masks = feature.data

                for i in range(self._dim - 1):
                    if i == len(sub_stack_masks):
                        sub_stack_masks.append([])
                    sub_stack_masks[i].append(sub_masks[i])

                unroll_features.append(sub_unroll_features)

            return unroll_features, [self._mask] + sub_stack_masks
